import logging
import re
from collections import defaultdict
from typing import Dict, List, Optional, Tuple
import platform

from gui_agents.core.knowledge import KnowledgeBase
from gui_agents.agents.global_state import GlobalState
from gui_agents.store.registry import Registry
from gui_agents.utils.common_utils import (
    Dag,
    Node,
    parse_dag,
    agent_log_to_string,
)
from gui_agents.tools.tools import Tools
from gui_agents.agents.stream_manager import stream_manager

logger = logging.getLogger("desktopenv.agent")

NUM_IMAGE_TOKEN = 1105  # Value set of screen of size 1920x1080 for openai vision

class Manager:
    def __init__(
        self,
        Tools_dict: Dict,
        # engine_params: Dict,
        local_kb_path: str,
        multi_round: bool = False,
        platform: str = platform.system().lower(),
        enable_search: bool = True,
    ):
        """
        Initialize the Manager which orchestrates planning, knowledge retrieval/fusion, DAG generation, topological sorting, and action queue creation for task-driven agents.
        
        Parameters:
            Tools_dict (Dict): Mapping of tool names to their configuration dictionaries; used to register and configure internal Tools instances (e.g., 'subtask_planner', 'dag_translator', 'embedding', 'websearch').
            local_kb_path (str): Filesystem path to the local knowledge base storage used by the KnowledgeBase.
            multi_round (bool): When True, enable multi-round interaction/stateful planning behavior across turns.
            platform (str): Target platform identifier (defaults to current system name); forwarded to KnowledgeBase and tools where applicable.
            enable_search (bool): When True, register and enable a web search tool ('websearch'); otherwise search functionality is disabled.
        """
        self.platform = platform
        self.Tools_dict = Tools_dict

        def _register(tools_instance, tool_name):
            """
            Register a tool with the provided tools manager using settings from Tools_dict.
            
            Parameters:
                tools_instance: An object exposing register_tool(name, provider, model, **kwargs) used to register the tool.
                tool_name (str): Key to look up the tool's configuration in Tools_dict; provider, model, and supported authentication keys
                    (e.g., 'api_key', 'base_url', 'endpoint_url', 'azure_endpoint', 'api_version') will be extracted and passed to register_tool.
            
            """
            config = Tools_dict.get(tool_name, {}).copy()
            provider = config.pop("provider", None)
            model = config.pop("model", None)

            auth_keys = ['api_key', 'base_url', 'endpoint_url', 'azure_endpoint', 'api_version']
            auth_params = {}
            for key in auth_keys:
                if key in config:
                    auth_params[key] = config[key]
                    logger.info(f"Manager._register: Setting {key} for tool '{tool_name}'")

            all_params = {**config, **auth_params}

            logger.info(f"Manager._register: Registering tool '{tool_name}' with provider '{provider}', model '{model}'")
            tools_instance.register_tool(tool_name, provider, model, **all_params)

        self.generator_agent = Tools()
        _register(self.generator_agent, "subtask_planner")

        self.dag_translator_agent = Tools()
        _register(self.dag_translator_agent, "dag_translator")

        self.narrative_summarization_agent = Tools()
        _register(self.narrative_summarization_agent, "narrative_summarization")

        self.episode_summarization_agent = Tools()
        _register(self.episode_summarization_agent, "episode_summarization")

        self.local_kb_path = local_kb_path

        self.embedding_engine = Tools()
        _register(self.embedding_engine, "embedding")
        
        KB_Tools_dict = {
            "embedding": self.Tools_dict.get("embedding"),
            "query_formulator": self.Tools_dict.get("query_formulator"),
            "context_fusion": self.Tools_dict.get("context_fusion"),
            "narrative_summarization": self.Tools_dict.get("narrative_summarization"),
            "episode_summarization": self.Tools_dict.get("episode_summarization"),
        }


        self.knowledge_base = KnowledgeBase(
            embedding_engine=self.embedding_engine,
            local_kb_path=self.local_kb_path,
            platform=platform,
            Tools_dict=KB_Tools_dict,
        )

        # GlobalState will be initialized in reset() method when task_id is available
        self.global_state: GlobalState = None  # type: ignore

        self.planner_history = []

        self.turn_count = 0
        self.task_id = None  # Will be set by agent

        # Initialize search engine based on enable_search parameter
        if enable_search:
            self.search_engine = Tools()
            _register(self.search_engine, "websearch")
        else:
            self.search_engine = None

        self.multi_round = multi_round

    def set_task_id(self, task_id: str) -> None:
        """Set the task identifier and update global state reference"""
        self.task_id = task_id
        # Update global state reference with task-specific registry
        self.global_state = Registry.get_from_context("GlobalStateStore", task_id)  # type: ignore

    def _send_stream_message(self, task_id: str, stage: str, message: str) -> None:
        """
        Enqueue a stream message for the given task if a task ID is provided.
        
        Parameters:
            task_id (str): Identifier of the task stream; no message is sent if empty.
            stage (str): Stage label for the message.
            message (str): Message content to enqueue.
        """
        if not task_id:
            return

        stream_manager.add_message_threadsafe(task_id, stage, message)

    def summarize_episode(self, trajectory):
        """
        Create a concise summary of the provided episode trajectory for lifelong learning and reflection.
        
        Parameters:
            trajectory (str): Serialized episode experience or trajectory to summarize.
        
        Returns:
            subtask_summarization (str): A short summary highlighting key subtasks, lessons, or reflections from the episode.
        """

        # Create Reflection on whole trajectories for next round trial, keep earlier messages as exemplars
        subtask_summarization, total_tokens, cost_string = self.episode_summarization_agent.execute_tool("episode_summarization", {"str_input": trajectory})
        logger.info(f"Episode summarization tokens: {total_tokens}, cost: {cost_string}")

        self.global_state.log_operation(
            module="manager",
            operation="episode_summarization",
            data={
                "tokens": total_tokens,
                "cost": cost_string,
                "content": subtask_summarization
            }
        )

        self._send_stream_message(self.task_id, "summarization", f"Episode summarization: {subtask_summarization}")

        return subtask_summarization

    def summarize_narrative(self, trajectory):
        """
        Produce a concise reflective summary of a narrative trajectory to inform lifelong learning.
        
        Parameters:
            trajectory: Narrative content (e.g., episode transcript or sequence of subtasks) to be summarized.
        
        Returns:
            A string containing a reflective summary that captures key insights, lessons learned, and recommendations for future rounds.
        """
        # Create Reflection on whole trajectories for next round trial
        lifelong_learning_reflection, total_tokens, cost_string = self.narrative_summarization_agent.execute_tool("narrative_summarization", {"str_input": trajectory})
        logger.info(f"Narrative summarization tokens: {total_tokens}, cost: {cost_string}")
        
        self.global_state.log_operation(
            module="manager",
            operation="narrative_summarization",
            data={
                "tokens": total_tokens,
                "cost": cost_string,
                "content": lifelong_learning_reflection
            }
        )

        return lifelong_learning_reflection
    
    def _generate_step_by_step_plan(
        self,
        observation: Dict,
        instruction: str,
        failed_subtask: Optional[Node] = None,
        completed_subtasks_list: List[Node] = [],
        remaining_subtasks_list: List[Node] = [],
    ) -> Tuple[Dict, str]:

        """
        Generate a high-level, step-by-step plan for the given task, optionally incorporating retrieved knowledge and the current subtask state.
        
        Parameters:
            observation (Dict): Current environment/desktop state; may include a 'screenshot' key with image data used for planning.
            instruction (str): Natural-language task description to plan for.
            failed_subtask (Optional[Node]): If provided, indicates a subtask that failed and triggers replanning for the remainder.
            completed_subtasks_list (List[Node]): Ordered list of subtasks already completed; used to inform replanning.
            remaining_subtasks_list (List[Node]): Ordered list of subtasks still expected; used to inform replanning.
        
        Returns:
            planner_info (Dict): Metadata about the planning step (includes at least 'search_query' and 'goal_plan').
            plan (str): The generated high-level plan as a human-readable string.
        
        Side effects:
            - May perform retrieval and knowledge fusion on the first planning turn.
            - Records operations to global_state, appends the plan to self.planner_history, increments self.turn_count, and sends stream messages when self.task_id is set.
        
        Raises:
            Exception: If plan generation produces an empty plan.
        """
        import time
        step_start = time.time()
        # Converts a list of DAG Nodes into a natural langauge list
        def format_subtask_list(subtasks: List[Node]) -> str:
            res = ""
            for idx, node in enumerate(subtasks):
                res += f"{idx+1}. **{node.name}**:\n"
                bullets = re.split(r"(?<=[.!?;]) +", node.info)
                for bullet in bullets:
                    res += f"   - {bullet}\n"
                res += "\n"
            return res
        prefix_message = ""
        # Perform Retrieval only at the first planning step
        if self.turn_count == 0:
            formulate_query_start = time.time()
            self.search_query, total_tokens, cost_string = self.knowledge_base.formulate_query(
                instruction, observation
            )
            formulate_query_time = time.time() - formulate_query_start
            logger.info(f"Formulate query tokens: {total_tokens}, cost: {cost_string}")
            self.global_state.log_operation(
                module="manager",
                operation="formulate_query",
                data={
                    "tokens": total_tokens,
                    "cost": cost_string,
                    "content": self.search_query,
                    "duration": formulate_query_time
                }
            )
            self.global_state.set_search_query(self.search_query)

            most_similar_task = ""
            retrieved_experience = ""
            integrated_knowledge = ""
            # Retrieve most similar narrative (task) experience
            narrative_start = time.time()
            most_similar_task, retrieved_experience, total_tokens, cost_string = (
                self.knowledge_base.retrieve_narrative_experience(instruction)
            )
            logger.info(f"Retrieve narrative experience tokens: {total_tokens}, cost: {cost_string}")
            narrative_time = time.time() - narrative_start
            logger.info(f"[Timing] Manager.retrieve_narrative_experience execution time: {narrative_time:.2f} seconds")
            self.global_state.log_operation(
                module="manager", 
                operation="retrieve_narrative_experience",
                data={
                    "tokens": total_tokens,
                    "cost": cost_string,
                    "content": "Most similar task: " + most_similar_task + "\n" + retrieved_experience.strip(),
                    "duration": narrative_time
                }
            )
            
            logger.info(
                "SIMILAR TASK EXPERIENCE: %s",
                most_similar_task + "\n" + retrieved_experience.strip(),
            )
            
            # Retrieve knowledge from the web if search_engine is provided
            if self.search_engine is not None:
                knowledge_start = time.time()
                retrieved_knowledge, total_tokens, cost_string = self.knowledge_base.retrieve_knowledge(
                    instruction=instruction,
                    search_query=self.search_query,
                    search_engine=self.search_engine,
                )
                logger.info(f"Retrieve knowledge tokens: {total_tokens}, cost: {cost_string}")
                
                knowledge_time = time.time() - knowledge_start
                logger.info(f"[Timing] Manager.retrieve_knowledge execution time: {knowledge_time:.2f} seconds")
                self.global_state.log_operation(
                    module="manager",
                    operation="retrieve_knowledge",
                    data={
                        "tokens": total_tokens,
                        "cost": cost_string,
                        "content": retrieved_knowledge,
                        "duration": knowledge_time
                    }
                )
                
                logger.info("RETRIEVED KNOWLEDGE: %s", retrieved_knowledge)

                if retrieved_knowledge is not None:
                    # Fuse the retrieved knowledge and experience
                    fusion_start = time.time()
                    integrated_knowledge, total_tokens, cost_string = self.knowledge_base.knowledge_fusion(
                        observation=observation,
                        instruction=instruction,
                        web_knowledge=retrieved_knowledge,
                        similar_task=most_similar_task,
                        experience=retrieved_experience,
                    )
                    logger.info(f"Knowledge fusion tokens: {total_tokens}, cost: {cost_string}")
                    fusion_time = time.time() - fusion_start
                    logger.info(f"[Timing] Manager.knowledge_fusion execution time: {fusion_time:.2f} seconds")
                    self.global_state.log_operation(
                        module="manager",
                        operation="knowledge_fusion",
                        data={
                            "tokens": total_tokens,
                            "cost": cost_string,
                            "content": integrated_knowledge,
                            "duration": fusion_time
                        }
                    )
                    
                    logger.info("INTEGRATED KNOWLEDGE: %s", integrated_knowledge)

            integrated_knowledge = integrated_knowledge or retrieved_experience

            # Add the integrated knowledge to the task instruction in the system prompt
            if integrated_knowledge:
                instruction += f"\nYou may refer to some retrieved knowledge if you think they are useful.{integrated_knowledge}"
            prefix_message = f"TASK_DESCRIPTION is {instruction}"

        # Re-plan on failure case
        if failed_subtask:
            agent_log = agent_log_to_string(self.global_state.get_agent_log())
            generator_message = (
                f"The subtask {failed_subtask} cannot be completed. Please generate a new plan for the remainder of the trajectory.\n\n"
                f"Successfully Completed Subtasks:\n{format_subtask_list(completed_subtasks_list)}\n"
                f"Please refer to the agent log to understand the progress and context of the task so far.\n{agent_log}"
            )
        # Re-plan on subtask completion case
        elif len(completed_subtasks_list) + len(remaining_subtasks_list) > 0:
            agent_log = agent_log_to_string(self.global_state.get_agent_log())
            generator_message = (
                "The current trajectory and desktop state is provided. Please revise the plan for the following trajectory.\n\n"
                f"Successfully Completed Subtasks:\n{format_subtask_list(completed_subtasks_list)}\n"
                f"Future Remaining Subtasks:\n{format_subtask_list(remaining_subtasks_list)}\n"
                f"Please refer to the agent log to understand the progress and context of the task so far.\n{agent_log}"
            )
        # Initial plan case
        else:
            generator_message = "Please generate the initial plan for the task.\n"
        
        generator_message = prefix_message + "\n" + generator_message
        logger.info("GENERATOR MESSAGE: %s", generator_message)
        logger.info("GENERATING HIGH LEVEL PLAN")

        subtask_planner_start = time.time()

        # Stream subtask planning message
        self._send_stream_message(self.task_id, "planning", "Analyzing tasks and generating subtask plans...")

        plan, total_tokens, cost_string = self.generator_agent.execute_tool("subtask_planner", {"str_input": generator_message, "img_input": observation.get("screenshot", None)})
        logger.info(f"Subtask planner tokens: {total_tokens}, cost: {cost_string}")
        subtask_planner_time = time.time() - subtask_planner_start
        logger.info(f"[Timing] Manager.subtask_planner execution time: {subtask_planner_time:.2f} seconds")
        self.global_state.log_operation(
            module="manager",
            operation="subtask_planner",
            data={
                "tokens": total_tokens,
                "cost": cost_string,
                "content": plan,
                "duration": subtask_planner_time
            }
        )

        # Stream planning completion message
        if self.task_id:
            plan_preview = plan[:150] + "..." if len(plan) > 150 else plan
            self._send_stream_message(self.task_id, "planning", f"Subtask planning completed: {plan_preview}")
        
        step_time = time.time() - step_start
        logger.info(f"[Timing] Manager._generate_step_by_step_plan execution time: {step_time:.2f} seconds")
        self.global_state.log_operation(
            module="manager",
            operation="Manager._generate_step_by_step_plan",
            data={"duration": step_time}
        )

        if plan == "":
            raise Exception("Plan Generation Failed - Fix the Prompt")

        logger.info("HIGH LEVEL STEP BY STEP PLAN: %s", plan)

        self.planner_history.append(plan)
        self.turn_count += 1

        planner_info = {
            "search_query": self.search_query,
            "goal_plan": plan,
        }

        assert type(plan) == str

        return planner_info, plan

    def _generate_dag(self, instruction: str, plan: str) -> Tuple[Dict, Dag]:
        import time
        dag_start = time.time()

        logger.info("GENERATING DAG")

        # Add maximum retry count
        max_retries = 2
        retry_count = 0
        dag = None

        while retry_count < max_retries and dag is None:
            if retry_count > 0:
                logger.warning(f"Retrying DAG generation, attempt {retry_count}")
                self.global_state.log_operation(
                    module="manager",
                    operation="dag_retry",
                    data={"retry_count": retry_count}
                )
            
            # Generate DAG
            dag_raw, total_tokens, cost_string = self.dag_translator_agent.execute_tool("dag_translator", {"str_input": f"Instruction: {instruction}\nPlan: {plan}"})
            logger.info(f"DAG translator tokens: {total_tokens}, cost: {cost_string}")

            # Try to parse DAG
            dag = parse_dag(dag_raw)
            
            # If parsing fails, increment retry count
            if dag is None:
                retry_count += 1
                # If not the last attempt, wait a short time before retrying
                if retry_count < max_retries:
                    time.sleep(1)
        
        dag_time = time.time() - dag_start
        logger.info(f"[Timing] Manager._generate_dag execution time: {dag_time:.2f} seconds")

        logger.info("Generated DAG: %s", dag_raw)
        self.global_state.log_operation(
            module="manager",
            operation="generated_dag",
            data={
                "tokens": total_tokens,
                "cost": cost_string,
                "content": dag_raw,
                "duration": dag_time,
                "retry_count": retry_count
            }
        )

        dag_info = {
            "dag": dag_raw,
        }

        # If all attempts fail, create a simple default DAG
        if dag is None:
            logger.error("Unable to generate valid DAG, using default DAG")
            # Create a simple default DAG with just one "Execute Task" node
            default_node = Node(name="Execute Task", info=f"Execute instruction: {instruction}")
            dag = Dag(nodes=[default_node], edges=[])
            
            self.global_state.log_operation(
                module="manager",
                operation="default_dag_created",
                data={"content": "Using default DAG because valid DAG could not be parsed from model output"}
            )

        return dag_info, dag

    def _topological_sort(self, dag: Dag) -> List[Node]:
        """Topological sort of the DAG using DFS
        dag: Dag: Object representation of the DAG with nodes and edges
        """
        import logging
        logger = logging.getLogger("desktopenv.agent")

        # Check if DAG is empty
        if not dag.nodes:
            logger.warning("DAG has no nodes, returning empty list")
            return []

        # If there's only one node, return it directly
        if len(dag.nodes) == 1:
            logger.info("DAG has only one node, returning directly")
            return dag.nodes

        def dfs(node_name, visited, temp_visited, stack):
            # If node is already in current path, we have a cycle
            if node_name in temp_visited:
                raise ValueError(f"Cycle detected in DAG involving node: {node_name}")
            
            # If node has been visited, skip
            if visited.get(node_name, False):
                return
            
            # Mark node as part of current path
            temp_visited.add(node_name)
            visited[node_name] = True
            
            # Visit all neighbors
            for neighbor in adj_list.get(node_name, []):
                if not visited.get(neighbor, False):
                    dfs(neighbor, visited, temp_visited, stack)
            
            # Remove node from current path
            temp_visited.remove(node_name)
            stack.append(node_name)

        try:
            # Build adjacency list
            adj_list = defaultdict(list)
            for u, v in dag.edges:
                if not u or not v:
                    logger.warning(f"Skipping invalid edge: {u} -> {v}")
                    continue
                adj_list[u.name].append(v.name)

            visited = {node.name: False for node in dag.nodes}
            temp_visited = set()  # For cycle detection
            stack = []

            # Perform DFS for each unvisited node
            for node in dag.nodes:
                if not visited.get(node.name, False):
                    dfs(node.name, visited, temp_visited, stack)

            # Return topologically sorted nodes
            sorted_nodes = []
            for name in stack[::-1]:
                matching_nodes = [n for n in dag.nodes if n.name == name]
                if matching_nodes:
                    sorted_nodes.append(matching_nodes[0])
                else:
                    logger.warning(f"Could not find node named {name} in DAG node list")
            
            # Check if all nodes are included in result
            if len(sorted_nodes) != len(dag.nodes):
                logger.warning(f"Number of nodes in topological sort result ({len(sorted_nodes)}) does not match number in DAG ({len(dag.nodes)})")
            
            return sorted_nodes
            
        except Exception as e:
            logger.error(f"Error during topological sort: {e}")
            # On error, return original node list
            logger.info("Returning unsorted original node list")
            return dag.nodes

    def get_action_queue(
        self,
        Tu: str,
        observation: Dict,
        running_state: str,
        failed_subtask: Optional[Node] = None,
        completed_subtasks_list: List[Node] = [],
        remaining_subtasks_list: List[Node] = [],
    ):
        """Generate the action list based on the instruction
        instruction:str: Instruction for the task
        """
        # Check for cancellation before starting action queue generation
        if self.global_state.is_cancelled():
            logger.info("Manager action queue generation cancelled by user request")
            return {"cancelled": True}, []

        import time
        action_queue_start = time.time()

        try:
            planner_info, plan = self._generate_step_by_step_plan(
                observation,
                Tu,
                failed_subtask,
                completed_subtasks_list,
                remaining_subtasks_list,
            )

            # Generate the DAG
            try:
                dag_info, dag = self._generate_dag(Tu, plan)
            except Exception as e:
                logger.error(f"Error generating DAG: {e}")
                # Create a simple default DAG with just one "Execute Task" node
                default_node = Node(name="Execute Task", info=f"Execute instruction: {Tu}")
                dag = Dag(nodes=[default_node], edges=[])
                dag_info = {"dag": "Failed to generate DAG, using default DAG"}
                
                self.global_state.log_operation(
                    module="manager",
                    operation="dag_generation_error",
                    data={"error": str(e), "content": "Using default DAG due to error in DAG generation"}
                )

            # Topological sort of the DAG
            try:
                action_queue = self._topological_sort(dag)
            except Exception as e:
                logger.error(f"Error during topological sort of DAG: {e}")
                # If topological sort fails, use node list directly
                action_queue = dag.nodes
                
                self.global_state.log_operation(
                    module="manager",
                    operation="topological_sort_error",
                    data={"error": str(e), "content": "Topological sort failed, using node list directly"}
                )

            planner_info.update(dag_info)
            
            if action_queue:
                logger.info(f"NEXT SUBTASK: {action_queue[0]}")
                self.global_state.log_operation(
                    module="manager",
                    operation="next_subtask",
                    data={"content": str(action_queue[0])}
                )
                
                if len(action_queue) > 1:
                    logger.info(f"REMAINING SUBTASKS: {action_queue[1:]}")
                    self.global_state.log_operation(
                        module="manager",
                        operation="remaining_subtasks",
                        data={"content": str(action_queue[1:])}
                    )
            
            action_queue_time = time.time() - action_queue_start
            logger.info(f"[Timing] manager.get_action_queue execution time: {action_queue_time:.2f} seconds")
            self.global_state.log_operation(
                module="manager",
                operation="manager.get_action_queue",
                data={"duration": action_queue_time}
            )

            return planner_info, action_queue
            
        except Exception as e:
            # Handle any unhandled exceptions in the entire process
            logger.error(f"Unhandled exception in get_action_queue function: {e}")
            
            # Create a simple default task node
            default_node = Node(name="Execute Task", info=f"Execute instruction: {Tu}")
            action_queue = [default_node]
            planner_info = {"error": str(e), "fallback": "Using default task node"}
            
            self.global_state.log_operation(
                module="manager",
                operation="get_action_queue_error",
                data={"error": str(e), "content": "Unhandled exception occurred, using default task node"}
            )
            
            action_queue_time = time.time() - action_queue_start
            logger.info(f"[Timing] manager.get_action_queue (error path) execution time: {action_queue_time:.2f} seconds")
            
            return planner_info, action_queue


#!/usr/bin/env python3
"""
Test script for Prometheus metrics integration.

This test verifies that the Prometheus metrics module works correctly with and without
prometheus_client installed.
"""

import sys
import os
from pathlib import Path

# Add the gui_agents/metrics directory directly to sys.path to avoid importing gui_agents.__init__
project_root = Path(__file__).parent.parent
metrics_path = project_root / "gui_agents" / "metrics"
sys.path.insert(0, str(metrics_path))


def test_metrics_without_prometheus():
    """Test that metrics work as no-ops when prometheus_client is not available."""
    print("=" * 80)
    print("Testing metrics without prometheus_client")
    print("=" * 80)
    
    # Force disable prometheus
    os.environ['ENABLE_PROMETHEUS'] = 'false'
    
    # Import directly from the module file
    import prometheus_metrics
    PrometheusMetrics = prometheus_metrics.PrometheusMetrics
    get_metrics_instance = prometheus_metrics.get_metrics_instance
    
    # Create metrics instance with prometheus disabled
    metrics = PrometheusMetrics(enabled=False)
    
    # Verify it's disabled
    assert not metrics.enabled, "Metrics should be disabled"
    
    # Test that all operations are no-ops (should not raise exceptions)
    try:
        metrics.record_task_created("pending")
        metrics.record_task_active(5)
        metrics.record_task_utilization(5, 10)
        metrics.record_task_execution_duration(120.5)
        metrics.record_task_queue_wait(5.2)
        
        metrics.record_grpc_request("RunAgentInstruction")
        metrics.record_grpc_error("RunAgentInstruction", "INTERNAL")
        metrics.record_grpc_stream_connection("GetAgentTaskStream", 1)
        metrics.record_grpc_stream_connection("GetAgentTaskStream", -1)
        
        metrics.record_tokens(1000, 200)
        metrics.record_cost(0.005, "USD")
        metrics.record_sandbox_created("Windows")
        metrics.record_task_steps(10)
        
        metrics.record_config_update("global")
        metrics.update_system_metrics(memory_bytes=1024*1024, temp_files=5, stream_tasks=3)
        metrics.update_success_rate(8, 10)
        metrics.update_uptime()
        
        # Test context manager
        with metrics.grpc_method_timer("TestMethod"):
            pass
        
        print("✅ All no-op operations completed without errors")
        return True
    except Exception as e:
        print(f"❌ No-op operations failed: {e}")
        import traceback
        traceback.print_exc()
        return False


def test_metrics_with_prometheus():
    """Test that metrics work correctly when prometheus_client is available."""
    print("\n" + "=" * 80)
    print("Testing metrics with prometheus_client")
    print("=" * 80)
    
    try:
        import prometheus_client
        print(f"✓ prometheus_client is available (version: {prometheus_client.__version__})")
    except ImportError:
        print("⚠ prometheus_client not installed, skipping this test")
        print("  Install with: pip install prometheus-client")
        return True  # Not a failure, just skipped
    
    # Enable prometheus
    os.environ['ENABLE_PROMETHEUS'] = 'true'
    
    import prometheus_metrics
    PrometheusMetrics = prometheus_metrics.PrometheusMetrics
    
    # Create metrics instance with prometheus enabled
    metrics = PrometheusMetrics(enabled=True)
    
    # Verify it's enabled
    assert metrics.enabled, "Metrics should be enabled"
    
    try:
        # Test task metrics
        metrics.record_task_created("pending")
        metrics.record_task_created("running")
        metrics.record_task_active(2)
        metrics.record_task_utilization(2, 5)
        metrics.record_task_execution_duration(45.5)
        metrics.record_task_queue_wait(2.3)
        
        # Test gRPC metrics
        metrics.record_grpc_request("RunAgentInstruction")
        metrics.record_grpc_request("QueryTaskStatus")
        metrics.record_grpc_error("RunAgentInstruction", "INTERNAL")
        metrics.record_grpc_stream_connection("GetAgentTaskStream", 1)
        metrics.record_grpc_stream_connection("GetAgentTaskStream", -1)
        
        # Test business metrics
        metrics.record_tokens(1500, 300)
        metrics.record_cost(0.008, "￥")
        metrics.record_sandbox_created("Windows")
        metrics.record_sandbox_created("Linux")
        metrics.record_task_steps(15)
        
        # Test system metrics
        metrics.record_config_update("global")
        metrics.update_system_metrics(memory_bytes=2048*1024, temp_files=10, stream_tasks=5)
        metrics.update_success_rate(18, 20)
        metrics.update_uptime()
        metrics.update_service_info(
            version="0.5.0",
            max_concurrent_tasks=5,
            log_level="INFO",
            domain="test-host"
        )
        
        # Test context manager
        with metrics.grpc_method_timer("TestMethod"):
            import time
            time.sleep(0.01)
        
        print("✅ All prometheus operations completed successfully")
        
        # Try to generate metrics output (won't start HTTP server in test)
        from prometheus_client import generate_latest
        metrics_output = generate_latest()
        
        # Verify some metrics are present in the output
        output_str = metrics_output.decode('utf-8')
        assert 'agent_tasks_created_total' in output_str, "Task creation metrics not found"
        assert 'agent_grpc_requests_total' in output_str, "gRPC request metrics not found"
        assert 'agent_tokens_consumed_total' in output_str, "Token metrics not found"
        
        print(f"✅ Metrics output generated ({len(output_str)} bytes)")
        print(f"   Sample metrics found in output:")
        print(f"   - agent_tasks_created_total")
        print(f"   - agent_grpc_requests_total")
        print(f"   - agent_tokens_consumed_total")
        
        return True
        
    except Exception as e:
        print(f"❌ Prometheus operations failed: {e}")
        import traceback
        traceback.print_exc()
        return False


def test_metrics_singleton():
    """Test that get_metrics_instance returns a singleton."""
    print("\n" + "=" * 80)
    print("Testing metrics singleton pattern")
    print("=" * 80)
    
    import prometheus_metrics
    get_metrics_instance = prometheus_metrics.get_metrics_instance
    
    # Reset environment
    os.environ['ENABLE_PROMETHEUS'] = 'false'
    
    # Get two instances
    metrics1 = get_metrics_instance()
    metrics2 = get_metrics_instance()
    
    # They should be the same object
    assert metrics1 is metrics2, "get_metrics_instance should return the same instance"
    
    print("✅ Singleton pattern verified")
    return True


def test_environment_variable_control():
    """Test that ENABLE_PROMETHEUS environment variable controls metrics."""
    print("\n" + "=" * 80)
    print("Testing environment variable control")
    print("=" * 80)
    
    import prometheus_metrics
    PrometheusMetrics = prometheus_metrics.PrometheusMetrics
    
    try:
        import prometheus_client
        prometheus_available = True
    except ImportError:
        prometheus_available = False
        print("⚠ prometheus_client not installed, testing with unavailable client")
    
    # Test with disabled
    os.environ['ENABLE_PROMETHEUS'] = 'false'
    metrics_disabled = PrometheusMetrics()
    assert not metrics_disabled.enabled, "Metrics should be disabled when ENABLE_PROMETHEUS=false"
    print("✅ ENABLE_PROMETHEUS=false correctly disables metrics")
    
    # Test with enabled (only works if prometheus_client is available)
    os.environ['ENABLE_PROMETHEUS'] = 'true'
    metrics_enabled = PrometheusMetrics()
    if prometheus_available:
        assert metrics_enabled.enabled, "Metrics should be enabled when ENABLE_PROMETHEUS=true and prometheus_client is available"
        print("✅ ENABLE_PROMETHEUS=true correctly enables metrics")
    else:
        assert not metrics_enabled.enabled, "Metrics should be disabled when prometheus_client is not available"
        print("✅ Metrics correctly disabled when prometheus_client not available")
    
    # Test with other truthy values
    for value in ['1', 'yes', 'True', 'TRUE']:
        os.environ['ENABLE_PROMETHEUS'] = value
        metrics_test = PrometheusMetrics()
        if prometheus_available:
            assert metrics_test.enabled, f"Metrics should be enabled with ENABLE_PROMETHEUS={value}"
    
    print("✅ All environment variable tests passed")
    return True


if __name__ == "__main__":
    success = True
    
    # Run all tests
    success &= test_metrics_without_prometheus()
    success &= test_metrics_with_prometheus()
    success &= test_metrics_singleton()
    success &= test_environment_variable_control()
    
    print("\n" + "=" * 80)
    if success:
        print("✅ All Prometheus metrics tests passed!")
        print("\nTo enable Prometheus monitoring in production:")
        print("  1. Install prometheus-client: pip install prometheus-client psutil")
        print("  2. Set environment variable: ENABLE_PROMETHEUS=true")
        print("  3. Optionally set PROMETHEUS_PORT=8000 (default)")
        print("  4. Configure Prometheus to scrape http://your-server:8000/metrics")
    else:
        print("❌ Some Prometheus metrics tests failed")
    print("=" * 80)
    
    sys.exit(0 if success else 1)



import base64

import numpy as np

from gui_agents.core.engine import (
    LMMEngineAnthropic,
    LMMEngineAzureOpenAI,
    LMMEngineHuggingFace,
    LMMEngineOpenAI,
    LMMEngineOpenRouter,
    LMMEnginevLLM,
    LMMEngineGemini,
    LMMEngineQwen,
    LMMEngineDoubao,
    LMMEngineDeepSeek,
    LMMEngineZhipu,
    LMMEngineGroq,
    LMMEngineSiliconflow,
    LMMEngineMonica,
    LMMEngineAWSBedrock,
    OpenAIEmbeddingEngine,
    GeminiEmbeddingEngine,
    AzureOpenAIEmbeddingEngine,
    DashScopeEmbeddingEngine,
    DoubaoEmbeddingEngine,
    JinaEmbeddingEngine,
    BochaAISearchEngine,
    ExaResearchEngine,
)

class CostManager:
    """Cost manager, responsible for adding currency symbols based on engine type"""
    
    # Chinese engines use CNY
    CNY_ENGINES = {
        LMMEngineQwen, LMMEngineDoubao, LMMEngineDeepSeek, LMMEngineZhipu, 
        LMMEngineSiliconflow, DashScopeEmbeddingEngine, DoubaoEmbeddingEngine
    }
    # Other engines use USD
    USD_ENGINES = {
        LMMEngineOpenAI, LMMEngineAnthropic, LMMEngineAzureOpenAI, LMMEngineGemini,
        LMMEngineOpenRouter, LMMEnginevLLM, LMMEngineHuggingFace, LMMEngineGroq,
        LMMEngineMonica, LMMEngineAWSBedrock, OpenAIEmbeddingEngine, 
        GeminiEmbeddingEngine, AzureOpenAIEmbeddingEngine, JinaEmbeddingEngine
    }
    
    @classmethod
    def get_currency_symbol(cls, engine) -> str:
        engine_type = type(engine)
        
        if engine_type in cls.CNY_ENGINES:
            return "￥"
        elif engine_type in cls.USD_ENGINES:
            return "$"
        else:
            return "$"
    
    @classmethod
    def format_cost(cls, cost: float, engine) -> str:
        currency = cls.get_currency_symbol(engine)
        return f"{cost:.7f}{currency}"
    
    @classmethod
    def add_costs(cls, cost1: str, cost2: str) -> str:
        currency_symbols = ["$", "￥", "¥", "€", "£"]
        currency1 = currency2 = "$"
        value1 = value2 = 0.0
        
        if isinstance(cost1, (int, float)):
            value1 = float(cost1)
            currency1 = "$"
        else:
            cost1_str = str(cost1)
            for symbol in currency_symbols:
                if symbol in cost1_str:
                    value1 = float(cost1_str.replace(symbol, "").strip())
                    currency1 = symbol
                    break
            else:
                try:
                    value1 = float(cost1_str)
                    currency1 = "$"
                except:
                    value1 = 0.0
        
        if isinstance(cost2, (int, float)):
            value2 = float(cost2)
            currency2 = "$"
        else:
            cost2_str = str(cost2)
            for symbol in currency_symbols:
                if symbol in cost2_str:
                    value2 = float(cost2_str.replace(symbol, "").strip())
                    currency2 = symbol
                    break
            else:
                try:
                    value2 = float(cost2_str)
                    currency2 = "$"
                except:
                    value2 = 0.0
        
        if currency1 != currency2:
            print(f"Warning: Different currencies in cost accumulation: {currency1} and {currency2}")
            currency = currency1
        else:
            currency = currency1
        
        total_value = value1 + value2
        return f"{total_value:.6f}{currency}"

class LLMAgent:
    def __init__(self, engine_params=None, system_prompt=None, engine=None):
        if engine is None:
            if engine_params is not None:
                engine_type = engine_params.get("engine_type")
                if engine_type == "openai":
                    self.engine = LMMEngineOpenAI(**engine_params)
                elif engine_type == "anthropic":
                    self.engine = LMMEngineAnthropic(**engine_params)
                elif engine_type == "azure":
                    self.engine = LMMEngineAzureOpenAI(**engine_params)
                elif engine_type == "vllm":
                    self.engine = LMMEnginevLLM(**engine_params)
                elif engine_type == "huggingface":
                    self.engine = LMMEngineHuggingFace(**engine_params)
                elif engine_type == "gemini":
                    self.engine = LMMEngineGemini(**engine_params)
                elif engine_type == "openrouter":
                    self.engine = LMMEngineOpenRouter(**engine_params)
                elif engine_type == "dashscope":
                    self.engine = LMMEngineQwen(**engine_params)
                elif engine_type == "doubao":
                    self.engine = LMMEngineDoubao(**engine_params)
                elif engine_type == "deepseek":
                    self.engine = LMMEngineDeepSeek(**engine_params)
                elif engine_type == "zhipu":
                    self.engine = LMMEngineZhipu(**engine_params)
                elif engine_type == "groq":
                    self.engine = LMMEngineGroq(**engine_params)
                elif engine_type == "siliconflow":
                    self.engine = LMMEngineSiliconflow(**engine_params)
                elif engine_type == "monica":
                    self.engine = LMMEngineMonica(**engine_params)
                elif engine_type == "aws_bedrock":
                    self.engine = LMMEngineAWSBedrock(**engine_params)
                else:
                    raise ValueError("engine_type is not supported")
            else:
                raise ValueError("engine_params must be provided")
        else:
            self.engine = engine

        self.messages = []  # Empty messages

        if system_prompt:
            self.add_system_prompt(system_prompt)
        else:
            self.add_system_prompt("You are a helpful assistant.")

    def encode_image(self, image_content):
        # if image_content is a path to an image file, check type of the image_content to verify
        if isinstance(image_content, str):
            with open(image_content, "rb") as image_file:
                return base64.b64encode(image_file.read()).decode("utf-8")
        else:
            return base64.b64encode(image_content).decode("utf-8")

    def reset(
        self,
    ):

        self.messages = [
            {
                "role": "system",
                "content": [{"type": "text", "text": self.system_prompt}],
            }
        ]

    def add_system_prompt(self, system_prompt):
        self.system_prompt = system_prompt
        if len(self.messages) > 0:
            self.messages[0] = {
                "role": "system",
                "content": [{"type": "text", "text": self.system_prompt}],
            }
        else:
            self.messages.append(
                {
                    "role": "system",
                    "content": [{"type": "text", "text": self.system_prompt}],
                }
            )

    def remove_message_at(self, index):
        """Remove a message at a given index"""
        if index < len(self.messages):
            self.messages.pop(index)

    def replace_message_at(
        self, index, text_content, image_content=None, image_detail="high"
    ):
        """Replace a message at a given index"""
        if index < len(self.messages):
            self.messages[index] = {
                "role": self.messages[index]["role"],
                "content": [{"type": "text", "text": text_content}],
            }
            if image_content:
                base64_image = self.encode_image(image_content)
                self.messages[index]["content"].append(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/png;base64,{base64_image}",
                            "detail": image_detail,
                        },
                    }
                )

    def add_message(
        self,
        text_content,
        image_content=None,
        role=None,
        image_detail="high",
        put_text_last=False,
    ):
        """Add a new message to the list of messages"""

        # API-style inference from OpenAI and similar services
        if isinstance(
            self.engine,
            (
                LMMEngineAnthropic,
                LMMEngineAzureOpenAI,
                LMMEngineHuggingFace,
                LMMEngineOpenAI,
                LMMEngineOpenRouter,
                LMMEnginevLLM,
                LMMEngineGemini,
                LMMEngineQwen,
                LMMEngineDoubao,
                LMMEngineDeepSeek,
                LMMEngineZhipu,
                LMMEngineGroq,
                LMMEngineSiliconflow,
                LMMEngineMonica,
                LMMEngineAWSBedrock,
            ),
        ):
            # infer role from previous message
            if role != "user":
                if self.messages[-1]["role"] == "system":
                    role = "user"
                elif self.messages[-1]["role"] == "user":
                    role = "assistant"
                elif self.messages[-1]["role"] == "assistant":
                    role = "user"

            message = {
                "role": role,
                "content": [{"type": "text", "text": text_content}],
            }

            if isinstance(image_content, np.ndarray) or image_content:
                # Check if image_content is a list or a single image
                if isinstance(image_content, list):
                    # If image_content is a list of images, loop through each image
                    for image in image_content:
                        base64_image = self.encode_image(image)
                        message["content"].append(
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": f"data:image/png;base64,{base64_image}",
                                    "detail": image_detail,
                                },
                            }
                        )
                else:
                    # If image_content is a single image, handle it directly
                    base64_image = self.encode_image(image_content)
                    message["content"].append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/png;base64,{base64_image}",
                                "detail": image_detail,
                            },
                        }
                    )

            # Rotate text to be the last message if desired
            if put_text_last:
                text_content = message["content"].pop(0)
                message["content"].append(text_content)

            self.messages.append(message)

        # For API-style inference from Anthropic
        elif isinstance(self.engine, (LMMEngineAnthropic, LMMEngineAWSBedrock)):
            # infer role from previous message
            if role != "user":
                if self.messages[-1]["role"] == "system":
                    role = "user"
                elif self.messages[-1]["role"] == "user":
                    role = "assistant"
                elif self.messages[-1]["role"] == "assistant":
                    role = "user"

            message = {
                "role": role,
                "content": [{"type": "text", "text": text_content}],
            }

            if image_content:
                # Check if image_content is a list or a single image
                if isinstance(image_content, list):
                    # If image_content is a list of images, loop through each image
                    for image in image_content:
                        base64_image = self.encode_image(image)
                        message["content"].append(
                            {
                                "type": "image",
                                "source": {
                                    "type": "base64",
                                    "media_type": "image/png",
                                    "data": base64_image,
                                },
                            }
                        )
                else:
                    # If image_content is a single image, handle it directly
                    base64_image = self.encode_image(image_content)
                    message["content"].append(
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": "image/png",
                                "data": base64_image,
                            },
                        }
                    )
            self.messages.append(message)

        # Locally hosted vLLM model inference
        elif isinstance(self.engine, LMMEnginevLLM):
            # infer role from previous message
            if role != "user":
                if self.messages[-1]["role"] == "system":
                    role = "user"
                elif self.messages[-1]["role"] == "user":
                    role = "assistant"
                elif self.messages[-1]["role"] == "assistant":
                    role = "user"

            message = {
                "role": role,
                "content": [{"type": "text", "text": text_content}],
            }

            if image_content:
                # Check if image_content is a list or a single image
                if isinstance(image_content, list):
                    # If image_content is a list of images, loop through each image
                    for image in image_content:
                        base64_image = self.encode_image(image)
                        message["content"].append(
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": f"data:image;base64,{base64_image}"
                                },
                            }
                        )
                else:
                    # If image_content is a single image, handle it directly
                    base64_image = self.encode_image(image_content)
                    message["content"].append(
                        {
                            "type": "image_url",
                            "image_url": {"url": f"data:image;base64,{base64_image}"},
                        }
                    )

            self.messages.append(message)
        else:
            raise ValueError("engine_type is not supported")

    def get_response(
        self,
        user_message=None,
        messages=None,
        temperature=0.0,
        max_new_tokens=None,
        **kwargs,
    ):
        """Generate the next response based on previous messages"""
        if messages is None:
            messages = self.messages
        if user_message:
            messages.append(
                {"role": "user", "content": [{"type": "text", "text": user_message}]}
            )

        content, total_tokens, cost = self.engine.generate(
            messages,
            temperature=temperature,
            max_new_tokens=max_new_tokens, # type: ignore
            **kwargs,
        )
        
        cost_string = CostManager.format_cost(cost, self.engine)
        
        return content, total_tokens, cost_string

class EmbeddingAgent:
    def __init__(self, engine_params=None, engine=None):
        if engine is None:
            if engine_params is not None:
                engine_type = engine_params.get("engine_type")
                if engine_type == "openai":
                    self.engine = OpenAIEmbeddingEngine(**engine_params)
                elif engine_type == "gemini":
                    self.engine = GeminiEmbeddingEngine(**engine_params)
                elif engine_type == "azure":
                    self.engine = AzureOpenAIEmbeddingEngine(**engine_params)
                elif engine_type == "dashscope":
                    self.engine = DashScopeEmbeddingEngine(**engine_params)
                elif engine_type == "doubao":
                    self.engine = DoubaoEmbeddingEngine(**engine_params)
                elif engine_type == "jina":
                    self.engine = JinaEmbeddingEngine(**engine_params)
                else:
                    raise ValueError(f"Embedding engine type '{engine_type}' is not supported")
            else:
                raise ValueError("engine_params must be provided")
        else:
            self.engine = engine

    def get_embeddings(self, text):
        """Get embeddings for the given text
        
        Args:
            text (str): The text to get embeddings for
            
        Returns:
            numpy.ndarray: The embeddings for the text
        """
        embeddings, total_tokens, cost = self.engine.get_embeddings(text)
        cost_string = CostManager.format_cost(cost, self.engine)
        return embeddings, total_tokens, cost_string

    
    def get_similarity(self, text1, text2):
        """Calculate the cosine similarity between two texts
        
        Args:
            text1 (str): First text
            text2 (str): Second text
            
        Returns:
            float: Cosine similarity score between the two texts
        """
        embeddings1, tokens1, cost1 = self.get_embeddings(text1)
        embeddings2, tokens2, cost2 = self.get_embeddings(text2)
        
        # Calculate cosine similarity
        dot_product = np.dot(embeddings1, embeddings2)
        norm1 = np.linalg.norm(embeddings1)
        norm2 = np.linalg.norm(embeddings2)
        
        similarity = dot_product / (norm1 * norm2)
        total_tokens = tokens1 + tokens2
        total_cost = CostManager.add_costs(cost1, cost2)
        
        return similarity, total_tokens, total_cost
    
    def batch_get_embeddings(self, texts):
        """Get embeddings for multiple texts
        
        Args:
            texts (List[str]): List of texts to get embeddings for
            
        Returns:
            List[numpy.ndarray]: List of embeddings for each text
        """
        embeddings = []
        total_tokens = [0, 0, 0]
        if texts:
            first_embedding, first_tokens, first_cost = self.get_embeddings(texts[0])
            embeddings.append(first_embedding)
            total_tokens[0] += first_tokens[0]
            total_tokens[1] += first_tokens[1]
            total_tokens[2] += first_tokens[2]
            total_cost = first_cost
            
            for text in texts[1:]:
                embedding, tokens, cost = self.get_embeddings(text)
                embeddings.append(embedding)
                total_tokens[0] += tokens[0]
                total_tokens[1] += tokens[1]
                total_tokens[2] += tokens[2]
                total_cost = CostManager.add_costs(total_cost, cost)
        else:
            currency = CostManager.get_currency_symbol(self.engine)
            total_cost = f"0.0{currency}"
        
        return embeddings, total_tokens, total_cost


class WebSearchAgent:
    def __init__(self, engine_params=None, engine=None):
        if engine is None:
            if engine_params is not None:
                self.engine_type = engine_params.get("engine_type")
                if self.engine_type == "bocha":
                    self.engine = BochaAISearchEngine(**engine_params)
                elif self.engine_type == "exa":
                    self.engine = ExaResearchEngine(**engine_params)
                else:
                    raise ValueError(f"Web search engine type '{self.engine_type}' is not supported")
            else:
                raise ValueError("engine_params must be provided")
        else:
            self.engine = engine
    
    def get_answer(self, query, **kwargs):
        """Get a direct answer for the query
        
        Args:
            query (str): The search query
            **kwargs: Additional arguments to pass to the search engine
            
        Returns:
            str: The answer text
        """
        if isinstance(self.engine, BochaAISearchEngine):
            answer, tokens, cost = self.engine.get_answer(query, **kwargs)
            return answer, tokens, str(cost)

        elif isinstance(self.engine, ExaResearchEngine):
            # For Exa, we'll use the chat_research method which returns a complete answer
            # results, tokens, cost = self.engine.search(query, **kwargs)
            results, tokens, cost = self.engine.chat_research(query, **kwargs)
            if isinstance(results, dict) and "messages" in results:
                for message in results.get("messages", []):
                    if message.get("type") == "answer":
                        return message.get("content", ""), tokens, str(cost)
            return str(results), tokens, str(cost)

        else:
            raise ValueError(f"Web search engine type '{self.engine_type}' is not supported")



# gui_agents/s2/store/registry.py

# Usage: in any file, get the object through Registry.get
# from gui_agents.store.registry import Registry
# GlobalStateStore = Registry.get("GlobalStateStore")

import threading
from typing import Optional, ClassVar


class Registry:
    """
    Registry class that supports both global singleton and task-specific instances.
    It uses a process-wide dictionary for task registries to ensure visibility
    across threads, making it compatible with asyncio.to_thread.
    """
    # For global singletons (backward compatibility)
    _global_services: ClassVar[dict[str, object]] = {}
    _global_lock: ClassVar[threading.RLock] = threading.RLock()

    # Process-wide storage for task-specific registries, protected by a lock
    _task_registries: ClassVar[dict[str, 'Registry']] = {}
    _task_registries_lock: ClassVar[threading.RLock] = threading.RLock()

    # Thread-local storage can be used as a cache for faster access
    _thread_local: ClassVar[threading.local] = threading.local()

    def __init__(self):
        """Create a new registry instance (for a specific task)."""
        self._services: dict[str, object] = {}
        self._lock = threading.RLock()

    # ========== Instance methods (for a single registry) ==========
    def register_instance(self, name: str, obj: object):
        """Register an object in this registry instance."""
        with self._lock:
            self._services[name] = obj

    def get_instance(self, name: str) -> object:
        """Get an object from this registry instance."""
        with self._lock:
            if name not in self._services:
                raise KeyError(f"{name!r} not registered in this Registry instance")
            return self._services[name]

    def clear_instance(self):
        """Clear all objects in this registry instance."""
        with self._lock:
            self._services.clear()

    # ========== Class methods for global registry (backward compatibility) ==========
    @classmethod
    def register(cls, name: str, obj: object):
        """Register an object in the global registry."""
        with cls._global_lock:
            cls._global_services[name] = obj

    @classmethod
    def get(cls, name: str) -> object:
        """Get an object from the global registry."""
        with cls._global_lock:
            if name not in cls._global_services:
                raise KeyError(f"{name!r} not registered in global Registry")
            return cls._global_services[name]

    @classmethod
    def clear(cls):
        """Clear all objects in the global registry."""
        with cls._global_lock:
            cls._global_services.clear()

    # ========== Task-specific registry management (Process-wide) ==========
    @classmethod
    def set_task_registry(cls, task_id: str, registry: 'Registry'):
        """Set a task-specific registry, making it visible process-wide."""
        with cls._task_registries_lock:
            cls._task_registries[task_id] = registry
        
        # Also set it in thread-local for faster access within the current thread
        if not hasattr(cls._thread_local, 'task_cache'):
            cls._thread_local.task_cache = {}
        cls._thread_local.task_cache[task_id] = registry

    @classmethod
    def get_task_registry(cls, task_id: str) -> Optional['Registry']:
        """Get a task-specific registry, checking thread-local cache first."""
        # Check thread-local cache first for performance
        if hasattr(cls._thread_local, 'task_cache'):
            cached_registry = cls._thread_local.task_cache.get(task_id)
            if cached_registry:
                return cached_registry

        # If not in cache, check the process-wide dictionary
        with cls._task_registries_lock:
            registry = cls._task_registries.get(task_id)
            if registry:
                # Populate cache for subsequent calls in the same thread
                if not hasattr(cls._thread_local, 'task_cache'):
                    cls._thread_local.task_cache = {}
                cls._thread_local.task_cache[task_id] = registry
            return registry

    @classmethod
    def remove_task_registry(cls, task_id: str):
        """Remove a task-specific registry from process-wide and thread-local storage."""
        # Remove from the main process-wide storage
        with cls._task_registries_lock:
            cls._task_registries.pop(task_id, None)
        
        # Remove from the current thread's local cache, if it exists
        if hasattr(cls._thread_local, 'task_cache'):
            cls._thread_local.task_cache.pop(task_id, None)

    @classmethod
    def get_from_context(cls, name: str, task_id: Optional[str] = None) -> object:
        """
        Get an object, trying task-specific registry first, then global registry.
        This is now thread-safe across different threads for the same task_id.
        """
        # Try task-specific registry first
        if task_id:
            task_registry = cls.get_task_registry(task_id)
            if task_registry:
                try:
                    return task_registry.get_instance(name)
                except KeyError:
                    pass  # Fall back to global registry

        # Fall back to global registry for CLI mode or if not in task registry
        return cls.get(name)



"""Core Agent Service implementation"""

import logging
import threading
import time
import uuid
import datetime
from concurrent.futures import ThreadPoolExecutor, Future
from typing import Dict, Optional, Any, Union
from pathlib import Path

from .api_models import (
    TaskRequest, TaskResult, TaskStatus, ExecutionStats, 
    AsyncTaskHandle, Backend, AgentMode
)
from .config import ServiceConfig
from .exceptions import (
    AgentServiceError, TaskExecutionError, TaskTimeoutError, 
    ConfigurationError, BackendError
)

# Import existing agent classes
from ..agents.agent_s import AgentS2, AgentSFast
from ..agents.hardware_interface import HardwareInterface
from ..store.registry import Registry
from ..agents.global_state import GlobalState

# Import backend classes for destroy_sandbox functionality
from ..agents.Backend.LybicBackend import LybicBackend
from ..agents.Backend.LybicMobileBackend import LybicMobileBackend



class AgentService:
    """
    Core service class that provides a unified interface for GUI automation tasks.
    
    This service wraps the existing Agent-S functionality and provides:
    - Synchronous and asynchronous task execution
    - Configuration management with multi-level API key support
    - Task lifecycle management
    - Execution statistics and monitoring
    """
    
    def __init__(
        self, 
        config: Optional[ServiceConfig] = None,
        **kwargs
    ):
        """
        Initialize the Agent Service
        
        Args:
            config: Service configuration. If None, will create from environment
            **kwargs: Override configuration parameters
        """
        # Initialize configuration
        if config is None:
            config = ServiceConfig.from_env()
        
        # Apply kwargs overrides
        for key, value in kwargs.items():
            if hasattr(config, key):
                setattr(config, key, value)
        
        # Validate configuration
        config.validate()
        
        self.config = config
        self.logger = self._setup_logging()
        
        # Task management
        self._tasks: Dict[str, TaskResult] = {}
        self._task_futures: Dict[str, Future] = {}
        self._task_lock = threading.RLock()
        
        # Thread pool for async execution
        self._executor = ThreadPoolExecutor(
            max_workers=config.max_concurrent_tasks,
            thread_name_prefix="AgentService"
        )
        
        # Agent instances cache
        self._agents: Dict[str, Union[AgentS2, AgentSFast]] = {}
        self._hwi_instances: Dict[str, HardwareInterface] = {}
        
        self.logger.info(f"AgentService initialized with config: {config.to_dict()}")
    
    def _setup_logging(self) -> logging.Logger:
        """Setup logging for the service"""
        logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
        logger.setLevel(getattr(logging, self.config.log_level.upper()))
        
        # Create log directory if it doesn't exist
        log_dir = Path(self.config.log_dir)
        log_dir.mkdir(parents=True, exist_ok=True)
        
        return logger
    
    def _get_or_create_agent(self, mode: str, task_id: Optional[str] = None, **kwargs) -> Union[AgentS2, AgentSFast]:
        """Get or create agent instance based on mode"""
        # Include task_id in cache key for task isolation when task_id is provided
        if task_id:
            cache_key = f"{mode}_{task_id}_{hash(str(sorted(kwargs.items())))}"
        else:
            cache_key = f"{mode}_{hash(str(sorted(kwargs.items())))}"

        if cache_key not in self._agents:
            agent_kwargs = {
                'platform': kwargs.get('platform', self.config.default_platform),
                'enable_takeover': kwargs.get('enable_takeover', self.config.enable_takeover),
                'enable_search': kwargs.get('enable_search', self.config.enable_search),
            }

            if mode == AgentMode.FAST.value:
                self._agents[cache_key] = AgentSFast(**agent_kwargs)
            else:
                self._agents[cache_key] = AgentS2(**agent_kwargs)

            self.logger.debug(f"Created new agent: {mode} with kwargs: {agent_kwargs}")

        # Set task_id on the agent for task-specific operations
        agent = self._agents[cache_key]
        if task_id and hasattr(agent, 'set_task_id'):
            agent.set_task_id(task_id)

        return agent
    
    def _get_or_create_hwi(self, backend: str, task_id: Optional[str] = None, **kwargs) -> HardwareInterface:
        """Get or create hardware interface instance"""
        if task_id:
            cache_key = f"{backend}_{task_id}_{hash(str(sorted(kwargs.items())))}"
        else:
            cache_key = f"{backend}_{hash(str(sorted(kwargs.items())))}"
        
        if cache_key not in self._hwi_instances:
            # Get backend-specific config
            backend_config = self.config.get_backend_config(backend)
            backend_config.update(kwargs)
            
            # Add platform info
            backend_config.setdefault('platform', self.config.default_platform)
            
            self._hwi_instances[cache_key] = HardwareInterface(
                backend=backend, 
                **backend_config
            )
            
            self.logger.debug(f"Created new HWI: {backend} with config: {backend_config}")
        
        return self._hwi_instances[cache_key]
    
    def _setup_global_state(self, task_id: str) -> str:
        """Setup global state for task execution with task isolation"""
        # Create timestamp-based directory structure like cli_app.py
        datetime_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        timestamp_dir = Path(self.config.log_dir) / f"{datetime_str}_{task_id[:8]}"  # Include task_id prefix
        cache_dir = timestamp_dir / "cache" / "screens"
        state_dir = timestamp_dir / "state"

        cache_dir.mkdir(parents=True, exist_ok=True)
        state_dir.mkdir(parents=True, exist_ok=True)

        # Create task-specific registry
        task_registry = Registry()

        # Register global state for this task in task-specific registry
        global_state = GlobalState(
            screenshot_dir=str(cache_dir),
            tu_path=str(state_dir / "tu.json"),
            search_query_path=str(state_dir / "search_query.json"),
            completed_subtasks_path=str(state_dir / "completed_subtasks.json"),
            failed_subtasks_path=str(state_dir / "failed_subtasks.json"),
            remaining_subtasks_path=str(state_dir / "remaining_subtasks.json"),
            termination_flag_path=str(state_dir / "termination_flag.json"),
            running_state_path=str(state_dir / "running_state.json"),
            display_info_path=str(timestamp_dir / "display.json"),
            agent_log_path=str(timestamp_dir / "agent_log.json")
        )

        # Register in task-specific registry using instance method
        registry_key = "GlobalStateStore"
        task_registry.register_instance(registry_key, global_state)

        # Set task registry in thread-local storage
        Registry.set_task_registry(task_id, task_registry)

        self.logger.info(f"Setup task-specific registry for task {task_id}")

        return str(timestamp_dir)
    
    def _execute_task_internal(self, request: TaskRequest, task_result: TaskResult) -> TaskResult:
        """Internal task execution method"""
        try:
            task_result.mark_started()
            self.logger.info(f"Starting task {task_result.task_id}: {request.instruction}")
            
            # Setup global state
            task_dir = self._setup_global_state(task_result.task_id)
            
            # Create agent and hardware interface with task_id
            agent = self._get_or_create_agent(
                request.mode,
                task_id=task_result.task_id,
                platform=self.config.default_platform,
                enable_takeover=request.enable_takeover,
                enable_search=request.enable_search
            )

            hwi = self._get_or_create_hwi(
                request.backend,
                task_id=task_result.task_id,
                **(request.config or {})
            )

            # Reset agent state
            agent.reset()
            
            # Execute task using existing run_agent logic
            start_time = time.time()
            
            if request.mode == AgentMode.FAST.value:
                self._run_agent_fast_internal(
                    agent, request.instruction, hwi, 
                    request.max_steps, request.enable_takeover,
                    request.destroy_sandbox, task_result.task_id
                )
            else:
                self._run_agent_normal_internal(
                    agent, request.instruction, hwi, 
                    request.max_steps, request.enable_takeover,
                    request.destroy_sandbox, task_result.task_id
                )
            
            end_time = time.time()
            
            # Create execution stats
            stats = ExecutionStats(
                total_duration=end_time - start_time,
                steps_count=0,  # Will be populated from global state if available
                tokens_used={"input": 0, "output": 0, "total": 0}
            )
            
            # Try to get more detailed stats from display.json
            try:
                display_json_path = Path(task_dir) / "display.json"
                if display_json_path.exists():
                    # Import here to avoid circular imports
                    # Use dynamic import to handle packaging issues
                    try:
                        from gui_agents.utils.analyze_display import analyze_display_json
                    except ImportError:
                        try:
                            from ..utils.analyze_display import analyze_display_json
                        except ImportError:
                            # Fallback for packaged version
                            import importlib
                            utils_module = importlib.import_module('gui_agents.utils')
                            analyze_display_json = getattr(utils_module.analyze_display, 'analyze_display_json')
                    analysis_result = analyze_display_json(str(display_json_path))
                    if analysis_result:
                        stats.steps_count = analysis_result.get('steps', 0)
                        stats.tokens_used = {
                            "input": analysis_result.get('input_tokens', 0),
                            "output": analysis_result.get('output_tokens', 0),
                            "total": analysis_result.get('total_tokens', 0)
                        }
                        stats.cost = analysis_result.get('cost', 0.0)
            except Exception as e:
                self.logger.warning(f"Failed to analyze execution stats: {e}")
            
            # Mark as completed
            task_result.mark_completed(
                result={"message": "Task completed successfully"},
                stats=stats
            )
            
            self.logger.info(
                f"Task {task_result.task_id} completed in {stats.total_duration:.2f}s "
                f"with {stats.steps_count} steps"
            )
            
        except Exception as e:
            error_msg = f"Task execution failed: {str(e)}"
            self.logger.error(error_msg, exc_info=True)
            task_result.mark_failed(error_msg)
            
        finally:
            # Cleanup task-specific registry
            Registry.remove_task_registry(task_result.task_id)
            self.logger.info(f"Cleaned up task-specific registry for task {task_result.task_id}")
        
        return task_result
    
    def _run_agent_normal_internal(self, agent, instruction: str, hwi, max_steps: int,
                                 enable_takeover: bool, destroy_sandbox: bool, task_id: str):
        """Run agent in normal mode (adapted from cli_app.py)"""
        # This is a simplified version - you may want to adapt the full logic from cli_app.py
        global_state: GlobalState = Registry.get_from_context("GlobalStateStore", task_id)  # type: ignore
        global_state.set_Tu(instruction)
        global_state.set_running_state("running")
        
        # Use dynamic import to handle packaging issues
        try:
            from gui_agents.agents.Action import Screenshot
        except ImportError:
            try:
                from ..agents.Action import Screenshot
            except ImportError:
                # Fallback for packaged version
                import importlib
                agents_module = importlib.import_module('gui_agents.agents')
                Screenshot = getattr(agents_module.Action, 'Screenshot')
        from PIL import Image
        
        try:
            for step in range(max_steps):
                # Take screenshot
                screenshot: Image.Image = hwi.dispatch(Screenshot())
                global_state.set_screenshot(screenshot)
                obs = global_state.get_obs_for_manager()
                
                # Get agent prediction
                info, code = agent.predict(instruction=instruction, observation=obs)
                
                # Check for completion
                if "done" in code[0]["type"].lower() or "fail" in code[0]["type"].lower():
                    agent.update_narrative_memory(f"Task: {instruction}")
                    break
                
                if "next" in code[0]["type"].lower():
                    continue
                    
                if "wait" in code[0]["type"].lower():
                    time.sleep(5)
                    continue
                
                # Execute action
                hwi.dispatchDict(code[0])
                time.sleep(1.0)
        finally:
            # Destroy sandbox if requested (only for Lybic backend)
            if destroy_sandbox:
                try:
                    if isinstance(hwi.backend, (LybicBackend, LybicMobileBackend)):
                        self.logger.info("Destroying sandbox as requested...")
                        hwi.backend.destroy_sandbox()
                except Exception as e:
                    self.logger.error(f"Failed to destroy sandbox: {e}")
    
    def _run_agent_fast_internal(self, agent, instruction: str, hwi, max_steps: int,
                              enable_takeover: bool, destroy_sandbox: bool, task_id: str):
        """Run agent in fast mode (adapted from cli_app.py)"""
        global_state: GlobalState = Registry.get_from_context("GlobalStateStore", task_id)  # type: ignore
        global_state.set_Tu(instruction)
        global_state.set_running_state("running")
        
        # Use dynamic import to handle packaging issues
        try:
            from gui_agents.agents.Action import Screenshot
        except ImportError:
            try:
                from ..agents.Action import Screenshot
            except ImportError:
                # Fallback for packaged version
                import importlib
                agents_module = importlib.import_module('gui_agents.agents')
                Screenshot = getattr(agents_module.Action, 'Screenshot')
        from PIL import Image
        
        try:
            for step in range(max_steps):
                # Take screenshot
                screenshot: Image.Image = hwi.dispatch(Screenshot())
                global_state.set_screenshot(screenshot)
                obs = global_state.get_obs_for_manager()
                
                # Get agent prediction
                info, code = agent.predict(instruction=instruction, observation=obs)
                
                # Check for completion
                if "done" in code[0]["type"].lower() or "fail" in code[0]["type"].lower():
                    break
                
                if "wait" in code[0]["type"].lower():
                    wait_duration = code[0].get("duration", 5000) / 1000
                    time.sleep(wait_duration)
                    continue
                
                # Execute action
                hwi.dispatchDict(code[0])
                time.sleep(0.5)
        finally:
            # Destroy sandbox if requested (only for Lybic backend)
            if destroy_sandbox:
                try:
                    if isinstance(hwi.backend, (LybicBackend, LybicMobileBackend)):
                        self.logger.info("[Fast Mode] Destroying sandbox as requested...")
                        hwi.backend.destroy_sandbox()
                except Exception as e:
                    self.logger.error(f"[Fast Mode] Failed to destroy sandbox: {e}")
    
    def execute_task(
        self, 
        instruction: str,
        backend: str | None = None,
        mode: str | None = None,
        max_steps: int | None = None,
        enable_takeover: bool | None = None,
        enable_search: bool | None = None,
        destroy_sandbox: bool | None = None,
        timeout: int | None = None,
        **kwargs
    ) -> TaskResult:
        """
        Execute a task synchronously
        
        Args:
            instruction: Task instruction in natural language
            backend: Backend to use (overrides config default)
            mode: Agent mode ('normal' or 'fast', overrides config default)
            max_steps: Maximum steps (overrides config default)
            enable_takeover: Enable user takeover (overrides config default)
            enable_search: Enable web search (overrides config default)
            destroy_sandbox: Destroy sandbox after task completion (overrides default: False)
            timeout: Task timeout in seconds (overrides config default)
            **kwargs: Additional configuration parameters
            
        Returns:
            TaskResult with execution details
        """
        # Create task request with defaults from config
        request = TaskRequest(
            instruction=instruction,
            backend=backend or self.config.default_backend,
            mode=mode or self.config.default_mode,
            max_steps=max_steps or self.config.default_max_steps,
            enable_takeover=enable_takeover if enable_takeover is not None else self.config.enable_takeover,
            enable_search=enable_search if enable_search is not None else self.config.enable_search,
            destroy_sandbox=destroy_sandbox if destroy_sandbox is not None else False,
            timeout=timeout or self.config.task_timeout,
            config=kwargs
        )
        
        # Create task result
        task_result = TaskResult.create_pending(instruction)
        
        # Store task
        with self._task_lock:
            self._tasks[task_result.task_id] = task_result
        
        # Execute task
        try:
            return self._execute_task_internal(request, task_result)
        finally:
            # Cleanup task future if exists
            with self._task_lock:
                self._task_futures.pop(task_result.task_id, None)
    
    def execute_task_async(
        self,
        instruction: str,
        **kwargs
    ) -> AsyncTaskHandle:
        """
        Execute a task asynchronously
        
        Args:
            instruction: Task instruction
            **kwargs: Same as execute_task
            
        Returns:
            AsyncTaskHandle for monitoring the task
        """
        # Create task request
        request = TaskRequest(
            instruction=instruction,
            backend=kwargs.get('backend', self.config.default_backend),
            mode=kwargs.get('mode', self.config.default_mode),
            max_steps=kwargs.get('max_steps', self.config.default_max_steps),
            enable_takeover=kwargs.get('enable_takeover', self.config.enable_takeover),
            enable_search=kwargs.get('enable_search', self.config.enable_search),
            destroy_sandbox=kwargs.get('destroy_sandbox', False),
            timeout=kwargs.get('timeout', self.config.task_timeout),
            config={k: v for k, v in kwargs.items() if k not in [
                'backend', 'mode', 'max_steps', 'enable_takeover', 
                'enable_search', 'destroy_sandbox', 'timeout'
            ]}
        )
        
        # Create task result
        task_result = TaskResult.create_pending(instruction)
        
        # Store task and submit to executor
        with self._task_lock:
            self._tasks[task_result.task_id] = task_result
            future = self._executor.submit(self._execute_task_internal, request, task_result)
            self._task_futures[task_result.task_id] = future
        
        return AsyncTaskHandle(task_id=task_result.task_id, status=TaskStatus.PENDING)
    
    def get_task_status(self, task_id: str) -> Optional[TaskResult]:
        """Get task status and result"""
        with self._task_lock:
            return self._tasks.get(task_id)
    
    def cancel_task(self, task_id: str) -> bool:
        """Cancel a running task"""
        with self._task_lock:
            # Cancel future if exists
            future = self._task_futures.get(task_id)
            if future:
                cancelled = future.cancel()
                if cancelled:
                    # Mark task as cancelled
                    task = self._tasks.get(task_id)
                    if task:
                        task.mark_cancelled()
                    return True
            return False
    
    def list_tasks(self, status: Optional[TaskStatus] = None) -> Dict[str, TaskResult]:
        """List all tasks, optionally filtered by status"""
        with self._task_lock:
            if status is None:
                return self._tasks.copy()
            else:
                return {
                    task_id: task for task_id, task in self._tasks.items()
                    if task.status == status
                }
    
    def cleanup_finished_tasks(self, max_age_seconds: int = 3600):
        """Clean up finished tasks older than max_age_seconds"""
        current_time = time.time()
        to_remove = []
        
        with self._task_lock:
            for task_id, task in self._tasks.items():
                if (task.is_finished and task.completed_at and 
                    current_time - task.completed_at > max_age_seconds):
                    to_remove.append(task_id)
            
            for task_id in to_remove:
                self._tasks.pop(task_id, None)
                self._task_futures.pop(task_id, None)
        
        if to_remove:
            self.logger.info(f"Cleaned up {len(to_remove)} finished tasks")
    
    def shutdown(self):
        """Shutdown the service and cleanup resources"""
        self.logger.info("Shutting down AgentService...")
        
        # Cancel all running tasks
        with self._task_lock:
            for task_id in list(self._task_futures.keys()):
                self.cancel_task(task_id)
        
        # Shutdown executor
        self._executor.shutdown(wait=True)
        
        # Clear caches
        self._agents.clear()
        self._hwi_instances.clear()
        
        self.logger.info("AgentService shutdown complete")
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown() 


import boto3
from botocore.exceptions import ClientError

import logging

from desktop_env.providers.base import Provider
from datetime import datetime
import time


logger = logging.getLogger("desktopenv.providers.aws.AWSProvider")
logger.setLevel(logging.INFO)

WAIT_DELAY = 15
MAX_ATTEMPTS = 10


class AWSProvider(Provider):


    def start_emulator(self, path_to_vm: str, headless: bool, *args, **kwargs):
        logger.info("Starting AWS VM...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            # Check the current state of the instance
            response = ec2_client.describe_instances(InstanceIds=[path_to_vm])
            state = response['Reservations'][0]['Instances'][0]['State']['Name']
            logger.info(f"Instance {path_to_vm} current state: {state}")

            if state == 'running':
                # If the instance is already running, skip starting it
                logger.info(f"Instance {path_to_vm} is already running. Skipping start.")
                return

            if state == 'stopped':
                # Start the instance if it's currently stopped
                ec2_client.start_instances(InstanceIds=[path_to_vm])
                logger.info(f"Instance {path_to_vm} is starting...")

                # Wait until the instance reaches 'running' state
                waiter = ec2_client.get_waiter('instance_running')
                waiter.wait(
                    InstanceIds=[path_to_vm],
                    WaiterConfig={'Delay': WAIT_DELAY, 'MaxAttempts': MAX_ATTEMPTS}
                )
                logger.info(f"Instance {path_to_vm} is now running.")
            else:
                # For all other states (terminated, pending, etc.), log a warning
                logger.warning(f"Instance {path_to_vm} is in state '{state}' and cannot be started.")

        except ClientError as e:
            logger.error(f"Failed to start the AWS VM {path_to_vm}: {str(e)}")
            raise


    def get_ip_address(self, path_to_vm: str) -> str:
        logger.info("Getting AWS VM IP address...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            response = ec2_client.describe_instances(InstanceIds=[path_to_vm])
            for reservation in response['Reservations']:
                for instance in reservation['Instances']:
                    private_ip_address = instance.get('PrivateIpAddress', '')
                    public_ip_address = instance.get('PublicIpAddress', '')
                    
                    if public_ip_address:
                        vnc_url = f"http://{public_ip_address}:5910/vnc.html"
                        logger.info("="*80)
                        logger.info(f"🖥️  VNC Web Access URL: {vnc_url}")
                        logger.info(f"📡 Public IP: {public_ip_address}")
                        logger.info(f"🏠 Private IP: {private_ip_address}")
                        logger.info("="*80)
                        print(f"\n🌐 VNC Web Access URL: {vnc_url}")
                        print(f"📍 Please open the above address in the browser for remote desktop access\n")
                    else:
                        logger.warning("No public IP address available for VNC access")
                    
                    return private_ip_address
            return ''  # Return an empty string if no IP address is found
        except ClientError as e:
            logger.error(f"Failed to retrieve IP address for the instance {path_to_vm}: {str(e)}")
            raise

    def save_state(self, path_to_vm: str, snapshot_name: str):
        logger.info("Saving AWS VM state...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            image_response = ec2_client.create_image(InstanceId=path_to_vm, Name=snapshot_name)
            image_id = image_response['ImageId']
            logger.info(f"AMI {image_id} created successfully from instance {path_to_vm}.")
            return image_id
        except ClientError as e:
            logger.error(f"Failed to create AMI from the instance {path_to_vm}: {str(e)}")
            raise

    def revert_to_snapshot(self, path_to_vm: str, snapshot_name: str):
        logger.info(f"Reverting AWS VM to snapshot AMI: {snapshot_name}...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            # Step 1: Retrieve the original instance details
            instance_details = ec2_client.describe_instances(InstanceIds=[path_to_vm])
            instance = instance_details['Reservations'][0]['Instances'][0]
            security_groups = [sg['GroupId'] for sg in instance['SecurityGroups']]
            subnet_id = instance['SubnetId']
            instance_type = instance['InstanceType']
            
            # Step 2: Terminate the old instance
            ec2_client.terminate_instances(InstanceIds=[path_to_vm])
            logger.info(f"Old instance {path_to_vm} has been terminated.")

            # Step 3: Launch a new instance from the snapshot(AMI) with performance optimization
            logger.info(f"Launching a new instance from AMI {snapshot_name}...")
            
            run_instances_params = {
                "MaxCount": 1,
                "MinCount": 1,
                "ImageId": snapshot_name,
                "InstanceType": instance_type,
                "EbsOptimized": True,
                "NetworkInterfaces": [
                    {
                        "SubnetId": subnet_id,
                        "AssociatePublicIpAddress": True,
                        "DeviceIndex": 0,
                        "Groups": security_groups
                    }
                ],
                "BlockDeviceMappings": [
                    {
                        "DeviceName": "/dev/sda1", 
                        "Ebs": {
                            # "VolumeInitializationRate": 300
                            "VolumeSize": 30,  # Size in GB
                            "VolumeType": "gp3",  # General Purpose SSD
                            "Throughput": 1000,
                            "Iops": 4000  # Adjust IOPS as needed
                        }
                    }
                ]
            }
            
            new_instance = ec2_client.run_instances(**run_instances_params)
            new_instance_id = new_instance['Instances'][0]['InstanceId']
            logger.info(f"New instance {new_instance_id} launched from AMI {snapshot_name}.")
            logger.info(f"Waiting for instance {new_instance_id} to be running...")
            ec2_client.get_waiter('instance_running').wait(InstanceIds=[new_instance_id])

            logger.info(f"Instance {new_instance_id} is ready.")
            
            try:
                instance_details = ec2_client.describe_instances(InstanceIds=[new_instance_id])
                instance = instance_details['Reservations'][0]['Instances'][0]
                public_ip = instance.get('PublicIpAddress', '')
                if public_ip:
                    vnc_url = f"http://{public_ip}:5910/vnc.html"
                    logger.info("="*80)
                    logger.info(f"🖥️  New Instance VNC Web Access URL: {vnc_url}")
                    logger.info(f"📡 Public IP: {public_ip}")
                    logger.info(f"🆔 New Instance ID: {new_instance_id}")
                    logger.info("="*80)
                    print(f"\n🌐 New Instance VNC Web Access URL: {vnc_url}")
                    print(f"📍 Please open the above address in the browser for remote desktop access\n")
            except Exception as e:
                logger.warning(f"Failed to get VNC address for new instance {new_instance_id}: {e}")

            return new_instance_id

        except ClientError as e:
            logger.error(f"Failed to revert to snapshot {snapshot_name} for the instance {path_to_vm}: {str(e)}")
            raise


    def stop_emulator(self, path_to_vm, region=None):
        logger.info(f"Stopping AWS VM {path_to_vm}...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            ec2_client.terminate_instances(InstanceIds=[path_to_vm])
            logger.info(f"Instance {path_to_vm} has been terminated.")
        except ClientError as e:
            logger.error(f"Failed to stop the AWS VM {path_to_vm}: {str(e)}")
            raise



import logging
import os
import subprocess
from typing import Dict
from xml.etree import ElementTree
from urllib.parse import urlparse

import acoustid
import cv2
import imagehash
import librosa
import numpy as np
from PIL import Image
from fastdtw import fastdtw
from scipy.spatial.distance import cosine
from skimage.metrics import structural_similarity as ssim

logger = logging.getLogger("desktopenv.metrics.vlc")


def is_vlc_playing(actual_status_path: str, rule: Dict[str, str]) -> float:
    """
    Checks if VLC is currently playing a file.
    """
    with open(actual_status_path, 'rb') as file:
        actual_status = file.read().decode('utf-8')

    tree = ElementTree.fromstring(actual_status)
    status = tree.find('state').text
    logger.info(f"VLC Status: {status}")
    if status == 'playing':
        if rule['type'] == 'file_name':
            # Try multiple possible paths for file information in VLC XML
            file_paths = [
                'information/category[@name="meta"]/info[@name="filename"]',
                'information/category[@name="meta"]/info[@name="title"]',
                'information/category[@name="meta"]/info[@name="uri"]',
                'information/category[@name="meta"]/info[@name="location"]',
                'information/category[@name="meta"]/info[@name="name"]'
            ]
            
            file_info = None
            for path in file_paths:
                element = tree.find(path)
                if element is not None and element.text:
                    file_info = element.text
                    break
            
            if file_info:
                expected_filename = rule['file_name']
                
                # Method 1: Direct filename match (most precise)
                actual_basename = os.path.basename(file_info)
                if actual_basename == expected_filename:
                    return 1
                
                # Method 2: Endswith match (for backward compatibility)
                if file_info.endswith(expected_filename):
                    return 1
                
                # Method 3: For paths, check if expected filename is in the path
                if expected_filename in file_info:
                    # Additional check to avoid false positives
                    # Make sure it's actually the filename, not just part of a path
                    if file_info.endswith('/' + expected_filename) or file_info.endswith('\\' + expected_filename):
                        return 1
                
                logger.warning(f"File name mismatch - Expected: {expected_filename}, Found: {file_info}")
                return 0
            else:
                logger.warning(f"Could not find file information in VLC status XML for rule: {rule}")
                return 0
        elif rule['type'] == 'url':
            # Try multiple possible paths for URL information in VLC XML
            url_paths = [
                'information/category[@name="meta"]/info[@name="url"]',
                'information/category[@name="meta"]/info[@name="URI"]', 
                'information/category[@name="meta"]/info[@name="location"]',
                'information/category[@name="meta"]/info[@name="title"]',  # Sometimes URL is in title for streams
                'information/category[@name="meta"]/info[@name="filename"]',  # Sometimes URL is in filename for streams
                'information/category[@name="Stream 0"]/info[@name="Codec"]',  # Try stream info
                'information/category[@name="Stream 0"]/info[@name="Type"]',
                'information/category[@name="Stream 0"]/info[@name="Language"]'
            ]
            
            file_info = None
            logger.debug(f"Looking for URL: {rule['url']}")
            
            for path in url_paths:
                element = tree.find(path)
                if element is not None and element.text:
                    file_info = element.text
                    logger.debug(f"Found URL info at '{path}': {file_info}")
                    break
            
            if file_info:
                # For URL comparison, check if the rule URL is contained in the file_info
                # This handles cases where VLC might show a longer or modified URL
                expected_url = rule['url']
                
                # Method 1: Direct URL match
                if expected_url in file_info or file_info.endswith(expected_url):
                    return 1
                
                # Method 2: For HLS streams, VLC often shows just the filename instead of full URL
                # Check if the file_info matches the filename part of the expected URL
                try:
                    expected_parsed = urlparse(expected_url)
                    expected_filename = os.path.basename(expected_parsed.path)
                    
                    # If VLC shows just the filename (common for HLS streams)
                    if file_info == expected_filename:
                        logger.info(f"URL filename match - Expected URL: {expected_url}, VLC shows filename: {file_info}")
                        return 1
                    
                    # Method 3: Check if both are URLs from the same domain and similar path
                    if '://' in file_info:  # file_info is also a URL
                        actual_parsed = urlparse(file_info)
                        # Same domain and similar path structure
                        if (expected_parsed.netloc == actual_parsed.netloc and 
                            expected_parsed.path in actual_parsed.path):
                            return 1
                except Exception as e:
                    logger.debug(f"URL parsing error: {e}")
                    pass
                
                logger.warning(f"URL mismatch - Expected: {expected_url}, Found: {file_info}")
                return 0
            else:
                logger.warning(f"Could not find URL information in VLC status XML for rule: {rule}")
                return 0
        else:
            logger.error(f"Unknown type: {rule['type']}")
            return 0
    else:
        return 0


# fixme: part of this function can be moved to getters
def is_vlc_recordings_folder(actual_config_path: str, rule: Dict[str, str]) -> float:
    """
    Checks if VLC's recording folder is set to the expected value.
    """
    with open(actual_config_path, 'rb') as file:
        config_file = file.read().decode('utf-8')

    expected_recording_file_path = rule['recording_file_path']

    try:
        for line in config_file.split("\n"):
            # Skip comments and empty lines
            if line.startswith('#') or not line.strip():
                continue
            # Check if the line contains the recording path setting
            if 'input-record-path' in line:
                # Extract the value of the recording path and remove surrounding whitespace
                current_path = line.split('=')[-1].strip()
                # Compare with the Desktop path
                if current_path == expected_recording_file_path:
                    return 1
                else:
                    return 0
        # The configuration key was not found in the file
        return 0
    except FileNotFoundError:
        logger.error("VLC configuration file not found.")
        return 0
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        return 0


def is_vlc_fullscreen(actual_window_size, screen_size):
    if screen_size is None or actual_window_size is None:
        # if the screen size is not available, means that the window is not fullscreen
        return 0

    if actual_window_size['width'] == screen_size['width'] and actual_window_size['height'] == screen_size['height']:
        return 1
    else:
        return 0


def compare_images(image1_path, image2_path, **options):
    # You would call this function with the paths to the two images you want to compare:
    # score = compare_images('path_to_image1', 'path_to_image2')
    # print("Similarity score:", score)

    if not image1_path or not image2_path:
        return 0

    base_score = options.get("reference_base_result", None)

    # Open the images and convert to grayscale
    image1 = Image.open(image1_path).convert('L')
    image2 = Image.open(image2_path).convert('L')

    # Resize images to the smaller one's size for comparison
    image1_size = image1.size
    image2_size = image2.size
    new_size = min(image1_size, image2_size)

    image1 = image1.resize(new_size, Image.Resampling.LANCZOS)
    image2 = image2.resize(new_size, Image.Resampling.LANCZOS)

    # Convert images to numpy arrays
    image1_array = np.array(image1)
    image2_array = np.array(image2)

    # Calculate SSIM between two images
    similarity_index = ssim(image1_array, image2_array)

    epsilon = 0.01
    if base_score is not None:
        if similarity_index >= base_score + epsilon:
            return (similarity_index - base_score) / (1 - base_score)
        else:
            return 0
    else:
        return similarity_index

def compare_audios(audio_path_1, audio_path_2):
    """
    Compare two audio files and return a similarity score in the range [0, 1].
    audio_path_1, audio_path_2: paths to the audio files to compare
    """
    # similarity = compare_audios_simple('path_to_audio1.mp3', 'path_to_audio2.mp3')
    # print(f'Similarity Score: {similarity}')

    if not audio_path_1 or not audio_path_2:
        return 0

    y1, y2 = None, None
    try:
        y1, sr1 = librosa.load(audio_path_1)
    except Exception:
        logger.warning(f"Could not load audio from {os.path.basename(audio_path_1)}. It might be empty or corrupt.")

    try:
        y2, sr2 = librosa.load(audio_path_2)
    except Exception:
        logger.warning(f"Could not load audio from {os.path.basename(audio_path_2)}. It might be empty or corrupt.")

    # Handle cases where one or both audio files are empty or corrupt.
    is_y1_bad = (y1 is None) or (y1.shape[0] == 0)
    is_y2_bad = (y2 is None) or (y2.shape[0] == 0)

    if is_y1_bad and is_y2_bad:
        logger.info("Both audio files are empty or corrupt. Considering them perfectly similar.")
        return 1.0
    
    if is_y1_bad or is_y2_bad:
        logger.warning(f"One audio file is empty/corrupt, the other is not. Similarity is 0.")
        return 0.0

    try:
        logger.info(f"Audio 1 ({os.path.basename(audio_path_1)}): sr={sr1}, len={len(y1)}")
        logger.info(f"Audio 2 ({os.path.basename(audio_path_2)}): sr={sr2}, len={len(y2)}")

        # Extract MFCC features
        mfcc1 = librosa.feature.mfcc(y=y1, sr=sr1)
        mfcc2 = librosa.feature.mfcc(y=y2, sr=sr2)
    except Exception as e:
        logger.error(f"Error during MFCC extraction: {e}")
        return 0.0

    # Normalize the MFCC features
    mfcc1 = librosa.util.normalize(mfcc1, axis=1)
    mfcc2 = librosa.util.normalize(mfcc2, axis=1)
    logger.info(f"MFCCs normalized.")

    # Define a lambda function to compute cosine distance
    dist_func = lambda x, y: cosine(x, y)

    # Use the DTW algorithm to find the best alignment path
    distance, path = fastdtw(mfcc1.T, mfcc2.T, dist=dist_func)
    logger.info(f"DTW distance: {distance:.4f}, Path length: {len(path)}")

    # Normalize the DTW distance by the length of the alignment path.
    if len(path) == 0:
        normalized_distance = np.inf
    else:
        normalized_distance = distance / len(path)
    logger.info(f"Normalized DTW distance: {normalized_distance:.4f}")
    
    # Convert the normalized distance to a similarity score using an exponential decay function.
    similarity = np.exp(-normalized_distance)

    return similarity


def compare_audios_by_dl_model(audio_path_1, audio_path_2):
    pass


def compare_videos(video_path1, video_path2, max_frames_to_check=100, threshold=5):
    # Open both video files
    cap1 = cv2.VideoCapture(video_path1)
    cap2 = cv2.VideoCapture(video_path2)

    frames_checked = 0
    mismatch_count = 0

    while frames_checked < max_frames_to_check:
        # Read frames from both videos
        ret1, frame1 = cap1.read()
        ret2, frame2 = cap2.read()

        # If a video ends, then check if both ended to confirm they are of the same length
        if not ret1 or not ret2:
            return 1. if ret1 == ret2 else 0. # return float only 

        # Convert frames to PIL Images
        frame1 = Image.fromarray(cv2.cvtColor(frame1, cv2.COLOR_BGR2RGB))
        frame2 = Image.fromarray(cv2.cvtColor(frame2, cv2.COLOR_BGR2RGB))

        # Compute the perceptual hash for each frame
        hash1 = imagehash.phash(frame1)
        hash2 = imagehash.phash(frame2)

        # Increment the frames checked
        frames_checked += 1

        # Compute the difference in the hashes
        if hash1 - hash2 > threshold:
            mismatch_count += 1
            # If there's a significant difference, the frames are not the same
            if mismatch_count > threshold:
                return 0.

    # If we reach here, the content appears to be the same
    return 1.


def check_qt_bgcone(actual_config_path, rule):
    with open(actual_config_path, 'rb') as file:
        config_file = file.read().decode('utf-8')

    expected_qt_bgcone = rule['expected_qt_bgcone']
    if isinstance(expected_qt_bgcone, int):
        expected_qt_bgcone = str(expected_qt_bgcone)

    try:
        # The default value of qt_bgcone is 1, which means it is enabled
        qt_bgcone = "1"
        for line in config_file.split("\n"):
            # Check if the line contains the recording path setting
            if 'qt-bgcone=' in line:
                # Extract the value of the recording path and remove surrounding whitespace
                qt_bgcone = line.split('=')[-1].strip()
            # The configuration key was not found in the file

        if qt_bgcone == expected_qt_bgcone:
            return 1
        else:
            return 0
    except FileNotFoundError:
        logger.error("VLC configuration file not found.")
        return 0
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        return 0


def check_qt_max_volume(actual_config_path, rule):
    with open(actual_config_path, 'rb') as file:
        config_file = file.read().decode('utf-8')

    expected_qt_max_volume = rule['expected_qt_max_volume']
    if isinstance(expected_qt_max_volume, int):
        expected_qt_max_volume = str(expected_qt_max_volume)

    try:
        qt_max_volume = "125"
        for line in config_file.split("\n"):
            if 'qt-max-volume=' in line:
                qt_max_volume = line.split('=')[-1].strip()
            # The configuration key was not found in the file

        if qt_max_volume == expected_qt_max_volume:
            return 1
        else:
            return 0
    except FileNotFoundError:
        logger.error("VLC configuration file not found.")
        return 0
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        return 0


def check_qt_minimal_view(actual_config_path, rule):
    with open(actual_config_path, 'rb') as file:
        config_file = file.read().decode('utf-8')

    expected_qt_minimal_view = rule['expected_qt_minimal_view']
    if isinstance(expected_qt_minimal_view, int):
        expected_qt_minimal_view = str(expected_qt_minimal_view)

    try:
        qt_minimal_view = "0"
        for line in config_file.split("\n"):
            if 'qt-minimal-view=' in line:
                qt_minimal_view = line.split('=')[-1].strip()

        if qt_minimal_view == expected_qt_minimal_view:
            return 1
        else:
            return 0
    except FileNotFoundError:
        logger.error("VLC configuration file not found.")
        return 0
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        return 0


def check_qt_slider_colours(actual_config_path, rule):
    with open(actual_config_path, 'rb') as file:
        config_file = file.read().decode('utf-8')

    try:
        qt_slider_colours = "153;210;153;20;210;20;255;199;15;245;39;29"
        for line in config_file.split("\n"):
            if 'qt-slider-colours' in line:
                qt_slider_colours = line.split('=')[-1].strip()
            # The configuration key was not found in the file

        if rule['type'] == 'match':
            expected_qt_slider_colours = rule['expected_qt_slider_colours']
            if qt_slider_colours == expected_qt_slider_colours:
                return 1
            else:
                return 0
        elif rule['type'] == 'blackish':
            def is_color_blackish(rgb_values, threshold=100):
                # decide if the color is blackish
                return all(value < threshold for value in rgb_values)

            def parse_qt_slider_colours(colours_string):
                # parse the string of colours into a list of RGB tuples
                values = [int(x) for x in colours_string.split(';')]
                colors = list(zip(values[0::3], values[1::3], values[2::3]))
                return colors

            colors = parse_qt_slider_colours(qt_slider_colours)

            # check if all colors are blackish
            for color in colors:
                if is_color_blackish(color):
                    pass
                else:
                    return 0
            return 1

    except FileNotFoundError:
        logger.error("VLC configuration file not found.")
        return 0
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        return 0


def check_global_key_play_pause(actual_config_path, rule):
    """
    # Play/Pause (str)
    #global-key-play-pause=

    # Play/Pause (str)
    #key-play-pause=Space
    """
    with open(actual_config_path, 'rb') as file:
        config_file = file.read().decode('utf-8')

    expected_global_key_play_pause = rule['expected_global_key_play_pause']

    if isinstance(expected_global_key_play_pause, int):
        expected_global_key_play_pause = str(expected_global_key_play_pause)

    try:
        global_key_play_pause = "0"
        for line in config_file.split("\n"):
            # Check if the line contains the recording path setting
            if 'global-key-play-pause=' in line:
                global_key_play_pause = "0" if line.split('=')[-1].strip() == "" else "1"

        if global_key_play_pause == expected_global_key_play_pause:
            return 1
        else:
            return 0
    except FileNotFoundError:
        logger.error("VLC configuration file not found.")
        return 0
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        return 0


def check_one_instance_when_started_from_file(actual_config_path, rule):
    with open(actual_config_path, 'rb') as file:
        config_file = file.read().decode('utf-8')

    expected_one_instance_when_started_from_file = rule['expected_one_instance_when_started_from_file']

    if isinstance(expected_one_instance_when_started_from_file, int):
        expected_one_instance_when_started_from_file = str(expected_one_instance_when_started_from_file)

    try:
        one_instance_when_started_from_file = "1"
        for line in config_file.split("\n"):
            # Check if the line contains the recording path setting
            if 'one-instance-when-started-from-file=' in line:
                one_instance_when_started_from_file = line.split('=')[-1].strip()

        if one_instance_when_started_from_file == expected_one_instance_when_started_from_file:
            return 1
        else:
            return 0
    except FileNotFoundError:
        logger.error("VLC configuration file not found.")
        return 0
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        return 0



# translator.py
"""
Translates pyautogui-style scripts into unified commands (JSON list),
format strictly follows computer-use schema.
"""

from __future__ import annotations
import ast
from typing import List, Dict


class TranslateError(RuntimeError):
    ...


class _CommandBuilder(ast.NodeVisitor):
    """
    Only handles the most common GUI atomic operations:
        click / moveTo / doubleClick / rightClick / middleClick /
        dragTo / scroll / typewrite / press / hotkey / wait
    If conditions, loops, or other logic are encountered, an error is thrown
    (the Grounding layer should flatten these first).
    """
    def __init__(self) -> None:
        super().__init__()
        self.cmds: List[Dict] = []

    # ---------- Node Visiting ----------
    def visit_Expr(self, node):         # pyautogui.xxx(...)
        if not isinstance(node.value, ast.Call):
            raise TranslateError("Only function call level instructions allowed")
        self._handle_call(node.value)
        self.generic_visit(node)

    # ---------- Core: Map function calls to commands ----------
    def _handle_call(self, call: ast.Call):
        if not isinstance(call.func, ast.Attribute):
            raise TranslateError("Complex expressions not supported")
        lib, fn = self._split_attr(call.func) # type: ignore
        if lib != "pyautogui":
            raise TranslateError("Only pyautogui calls allowed")

        # Get positional and keyword arguments
        kw = {k.arg: self._literal(v) for k, v in zip(call.keywords, [k.value for k in call.keywords])}
        pos = [self._literal(a) for a in call.args]

        # ---------- mapping ----------
        if fn in {"click", "doubleClick", "rightClick", "middleClick"}:
            x, y = pos[:2] if len(pos) >= 2 else (kw.get("x"), kw.get("y"))
            self._append_click(fn, x, y, kw)

        elif fn == "moveTo":
            x, y = pos[:2] if len(pos) >= 2 else (kw.get("x"), kw.get("y"))
            self.cmds.append({"action": "move", "coordinate": [x, y]})

        elif fn == "dragTo":
            x, y = pos[:2] if len(pos) >= 2 else (kw.get("x"), kw.get("y"))
            # startCoordinate needs to be supplemented by the caller; using None as a placeholder here
            self.cmds.append({"action": "leftClickDrag",
                              "startCoordinate": None,
                              "coordinate": [x, y]})

        elif fn == "scroll":
            clicks = pos[0] if pos else kw.get("clicks")
            direction = "up" if clicks > 0 else "down"
            coordinate = [kw.get("x", 0), kw.get("y", 0)]
            self.cmds.append({"action": "scroll",
                              "scrollAmount": abs(clicks),
                              "scrollDirection": direction,
                              "coordinate": coordinate})

        elif fn in {"typewrite", "write"}:
            text = pos[0] if pos else kw.get("message")
            self.cmds.append({"action": "type", "text": text})

        elif fn in {"press", "hotkey"}:
            keys = [self._literal(a) for a in call.args]
            key_combo = "+".join(keys)
            self.cmds.append({"action": "keyPress", "text": key_combo})

        elif fn == "sleep":
            secs = pos[0] if pos else kw.get("seconds", 1)
            self.cmds.append({"action": "wait", "duration": secs})

        else:
            raise TranslateError(f"Function {fn} not yet supported")

    # ---------- Tools ----------
    def _append_click(self, fn, x, y, kw):
        # Single click / double click / different buttons
        clicks = kw.get("clicks", 1)
        button = kw.get("button", "left")
        action = {
            ("click", 1, "left"): "click",
            ("click", 2, "left"): "doubleClick",
            ("doubleClick", 1, "left"): "doubleClick",
            ("rightClick", 1, "right"): "rightClick",
            ("middleClick", 1, "middle"): "middleClick",
        }.get((fn, clicks, button))
        if not action:
            raise TranslateError(f"Cannot map {fn} clicks={clicks} button={button}")
        self.cmds.append({"action": action, "coordinate": [x, y]})

    def _split_attr(self, attr: ast.Attribute):
        parts = []
        while isinstance(attr, ast.Attribute):
            parts.insert(0, attr.attr)
            attr = attr.value # type: ignore
            if isinstance(attr, ast.Name):
                parts.insert(0, attr.id)
            else:
                raise TranslateError("Complex expressions not supported")
        return parts[0], parts[1]

    def _literal(self, node):
        if isinstance(node, ast.Constant):
            return node.value
        raise TranslateError("Only literal parameters allowed")

# ---------- External API ----------
def translate(py_code: str) -> List[Dict]:
    tree = ast.parse(py_code)
    builder = _CommandBuilder()
    builder.visit(tree)
    return builder.cmds


# ---------------- demo ----------------
# if __name__ == "__main__":
#     sample = "import pyautogui; pyautogui.click(769, 1006, clicks=1, button='left');"
#     cmds = translate(sample)
#     print(json.dumps(cmds, indent=2, ensure_ascii=False))



import unittest
import os
import json
import logging
import sys
from io import BytesIO

from gui_agents.agents.manager import Manager
from gui_agents.utils.common_utils import Node, Dag

# Configure colored logging
class ColoredFormatter(logging.Formatter):
    """Custom colored logging formatter"""
    COLORS = {
        'DEBUG': '\033[94m',  # Blue
        'INFO': '\033[92m',   # Green
        'WARNING': '\033[93m', # Yellow
        'ERROR': '\033[91m',  # Red
        'CRITICAL': '\033[91m\033[1m', # Red bold
        'RESET': '\033[0m'    # Reset
    }
    
    def format(self, record):
        log_message = super().format(record)
        return f"{self.COLORS.get(record.levelname, self.COLORS['RESET'])}{log_message}{self.COLORS['RESET']}"

# Configure logging - Clear all handlers and reconfigure
logger = logging.getLogger(__name__)
logger.handlers = []  # Clear all existing handlers
logger.propagate = False  # Prevent logging from propagating to root logger

# Add single handler
console_handler = logging.StreamHandler(sys.stdout)
console_handler.setFormatter(ColoredFormatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
logger.addHandler(console_handler)
logger.setLevel(logging.INFO)

# Define colored separator
def print_test_header(test_name):
    """Print test header, using colored and prominent separator"""
    separator = "="*80
    logger.info(separator)
    logger.info(test_name.center(80))
    logger.info(separator)

def print_test_section(section_name):
    """Print test section, using colored and prominent separator"""
    separator = "-"*60
    logger.info("\n" + separator)
    logger.info(section_name.center(60))
    logger.info(separator)

class TestManager(unittest.TestCase):
    def setUp(self):
        """Set up test environment"""
        print_test_header("Set up test environment")
        
        # Load tools configuration file
        tools_config_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "tools", "tools_config.json")
        with open(tools_config_path, "r") as f:
            tools_config = json.load(f)
            self.Tools_dict = {}
            for tool in tools_config["tools"]:
                tool_name = tool["tool_name"]
                self.Tools_dict[tool_name] = {
                    "provider": tool["provider"],
                    "model": tool["model_name"]
                }
        logger.info(f"Loaded {len(self.Tools_dict)} tool configurations")
        
        # Create test directory structure
        self.test_kb_path = "test_kb"
        self.platform = "darwin"
        self.test_platform_path = os.path.join(self.test_kb_path, self.platform)
        os.makedirs(self.test_platform_path, exist_ok=True)
        logger.info(f"Created test directory: {self.test_platform_path}")
        
        # Create test files
        with open(os.path.join(self.test_platform_path, "narrative_memory.json"), "w") as f:
            json.dump({}, f)
        with open(os.path.join(self.test_platform_path, "episodic_memory.json"), "w") as f:
            json.dump({}, f)
        with open(os.path.join(self.test_platform_path, "embeddings.pkl"), "wb") as f:
            f.write(b"")
        logger.info("Created test files")
        
        # Create Manager instance - use actual Manager instead of mock
        self.manager = Manager(
            Tools_dict=self.Tools_dict,
            local_kb_path=self.test_kb_path,
            platform=self.platform
        )
        logger.info("Manager instance created")
        
        # Create test observation data
        import pyautogui
        self.test_image = pyautogui.screenshot()
        buffered = BytesIO()
        self.test_image.save(buffered, format="PNG")
        self.test_screenshot_bytes = buffered.getvalue()
        
        self.test_observation = {
            "screenshot": self.test_screenshot_bytes
        }
        logger.info("Test observation data created")
        
        # Test instruction
        self.test_instruction = "在系统中打开设置并更改显示分辨率"
        logger.info(f"Test instruction: {self.test_instruction}")

    def tearDown(self):
        """Clean up test environment"""
        print_test_header("Clean up test environment")
        import shutil
        if os.path.exists(self.test_kb_path):
            shutil.rmtree(self.test_kb_path)
        logger.info(f"Deleted test directory: {self.test_kb_path}")

    def test_generate_step_by_step_plan(self):
        """Test _generate_step_by_step_plan method"""
        print_test_header("Test _generate_step_by_step_plan method")
        logger.info(f"Input parameters: observation={type(self.test_observation)}, instruction={self.test_instruction}")
        
        # Test initial plan generation
        print_test_section("Initial plan generation")
        planner_info, plan = self.manager._generate_step_by_step_plan(
            self.test_observation,
            self.test_instruction
        )
        
        # Output results
        logger.info(f"Output results: planner_info={planner_info}")
        logger.info(f"Output results: plan(first 100 characters)={plan[:100]}...")
        
        # Verify results
        self.assertIsNotNone(plan)
        self.assertIsInstance(plan, str)
        self.assertGreater(len(plan), 0)
        self.assertIn("search_query", planner_info)
        self.assertIn("goal_plan", planner_info)
        self.assertEqual(planner_info["goal_plan"], plan)
        
        # Test re-planning (failed subtask)
        print_test_section("Test re-planning (failed subtask)")
        failed_subtask = Node(name="Failed subtask", info="Failed subtask information")
        completed_subtasks = [Node(name="Completed subtask", info="Completed subtask information")]
        
        logger.info(f"Input parameters: failed_subtask={failed_subtask}, completed_subtasks={completed_subtasks}")
        
        self.manager.turn_count = 1  # Set to non-initial state
        planner_info, plan = self.manager._generate_step_by_step_plan(
            self.test_observation,
            self.test_instruction,
            failed_subtask,
            completed_subtasks,
            []
        )
        
        # Output results
        logger.info(f"Output results: planner_info={planner_info}")
        logger.info(f"Output results: plan(first 100 characters)={plan[:100]}...")
        
        # Verify results
        self.assertIsNotNone(plan)
        self.assertIsInstance(plan, str)
        self.assertGreater(len(plan), 0)
        self.assertIn("goal_plan", planner_info)
        self.assertEqual(planner_info["goal_plan"], plan)

    def test_generate_dag(self):
        """Test _generate_dag method"""
        print_test_header("Test _generate_dag method")
        
        # First generate plan
        print_test_section("Generate plan")
        logger.info("First generate plan")
        _, plan = self.manager._generate_step_by_step_plan(
            self.test_observation,
            self.test_instruction
        )
        logger.info(f"Generated plan(first 100 characters): {plan[:100]}...")
        
        # Use generated plan to create DAG
        print_test_section("Create DAG")
        logger.info(f"Input parameters: instruction={self.test_instruction}, plan(first 100 characters)={plan[:100]}...")
        dag_raw = self.manager.dag_translator_agent.execute_tool("dag_translator", {"str_input": f"Instruction: {self.test_instruction}\nPlan: {plan}"})
        logger.info(f"Raw DAG output: {dag_raw}")
        
        # Manually parse DAG
        print_test_section("Parse DAG")
        from gui_agents.utils.common_utils import parse_dag
        dag = parse_dag(dag_raw)
        
        if dag is None:
            logger.error("DAG parsing failed, create a simple test DAG")
            # Create a simple test DAG
            nodes = [
                Node(name="Open settings", info="Open settings application"),
                Node(name="Navigate to display settings", info="Find and click on display settings option"),
                Node(name="Change resolution", info="Change screen resolution")
            ]
            edges = [
                [nodes[0], nodes[1]],
                [nodes[1], nodes[2]]
            ]
            dag = Dag(nodes=nodes, edges=edges)
        
        dag_info = {"dag": dag_raw}
        
        logger.info(f"Parsed DAG: nodes={[node.name for node in dag.nodes]}, edges number={len(dag.edges)}")
        
        # Verify results
        self.assertIsNotNone(dag)
        self.assertIsInstance(dag, Dag)
        self.assertGreater(len(dag.nodes), 0)
        self.assertGreaterEqual(len(dag.edges), 0)
        self.assertIn("dag", dag_info)

    def test_topological_sort(self):
        """Test _topological_sort method"""
        print_test_header("Test _topological_sort method")
        
        # Create test DAG
        print_test_section("Create test DAG")
        nodes = [
            Node(name="A", info="Task A"),
            Node(name="B", info="Task B"),
            Node(name="C", info="Task C"),
            Node(name="D", info="Task D")
        ]
        
        edges = [
            [nodes[0], nodes[1]],  # A -> B
            [nodes[0], nodes[2]],  # A -> C
            [nodes[1], nodes[3]],  # B -> D
            [nodes[2], nodes[3]]   # C -> D
        ]
        
        dag = Dag(nodes=nodes, edges=edges)
        logger.info(f"Input parameters: dag.nodes={[node.name for node in dag.nodes]}, dag.edges number={len(dag.edges)}")
        
        # Execute topological sort
        print_test_section("Execute topological sort")
        sorted_nodes = self.manager._topological_sort(dag)
        logger.info(f"Output results: sorted_nodes={[node.name for node in sorted_nodes]}")
        
        # Verify results
        print_test_section("Verify sorting results")
        self.assertEqual(len(sorted_nodes), 4)
        self.assertEqual(sorted_nodes[0].name, "A")
        
        # Verify B and C's order may be uncertain, but they are both after A and before D
        self.assertIn(sorted_nodes[1].name, ["B", "C"])
        self.assertIn(sorted_nodes[2].name, ["B", "C"])
        self.assertNotEqual(sorted_nodes[1].name, sorted_nodes[2].name)
        
        self.assertEqual(sorted_nodes[3].name, "D")

    def test_get_action_queue(self):
        """Test get_action_queue method"""
        print_test_header("Test get_action_queue method")
        
        # Modify Manager's _generate_dag method to avoid parsing failure
        print_test_section("Modify _generate_dag method")
        def mock_generate_dag(self, instruction, plan):
            logger.info("Use modified _generate_dag method")
            dag_raw = self.dag_translator_agent.execute_tool("dag_translator", {"str_input": f"Instruction: {instruction}\nPlan: {plan}"})
            logger.info(f"Raw DAG output: {dag_raw}")
            
            # Try to parse DAG
            from gui_agents.utils.common_utils import parse_dag
            dag = parse_dag(dag_raw)
            
            # If parsing fails, create a simple test DAG
            if dag is None:
                logger.warning("DAG parsing failed, create a simple test DAG")
                nodes = [
                    Node(name="Open settings", info="Open settings application"),
                    Node(name="Navigate to display settings", info="Find and click on display settings option"),
                    Node(name="Change resolution", info="Change screen resolution")
                ]
                edges = [
                    [nodes[0], nodes[1]],
                    [nodes[1], nodes[2]]
                ]
                dag = Dag(nodes=nodes, edges=edges)
            
            dag_info = {"dag": dag_raw}
            return dag_info, dag
        
        # Replace original method
        original_generate_dag = self.manager._generate_dag
        self.manager._generate_dag = lambda instruction, plan: mock_generate_dag(self.manager, instruction, plan)
        
        try:
            # Call get_action_queue method
            print_test_section("Call get_action_queue method")
            logger.info(f"Input parameters: Tu={self.test_instruction}, Screenshot=Image(100x100), Running_state='初始状态'")
            planner_info, action_queue = self.manager.get_action_queue(
                Tu=self.test_instruction,
                Screenshot=self.test_image,
                Running_state="初始状态"
            )
            
            # Output results
            print_test_section("Verify results")
            logger.info(f"Output results: planner_info={planner_info}")
            logger.info(f"Output results: action_queue={[action.name for action in action_queue]}")
            
            # Verify results
            self.assertIsNotNone(planner_info)
            self.assertIsNotNone(action_queue)
            self.assertIn("search_query", planner_info)
            self.assertIn("goal_plan", planner_info)
            self.assertIn("dag", planner_info)
            self.assertGreater(len(action_queue), 0)
            
            # Verify that the elements in action_queue are Node types
            for action in action_queue:
                self.assertIsInstance(action, Node)
                self.assertIsNotNone(action.name)
                self.assertIsNotNone(action.info)
        finally:
            # Restore original method
            self.manager._generate_dag = original_generate_dag

if __name__ == '__main__':
    unittest.main()


import boto3
from botocore.exceptions import ClientError
import base64
import logging
import json
from typing import Optional

from desktop_env.providers.base import Provider
from desktop_env.providers.aws.proxy_pool import get_global_proxy_pool, init_proxy_pool, ProxyInfo

logger = logging.getLogger("desktopenv.providers.aws.AWSProviderWithProxy")
logger.setLevel(logging.INFO)

WAIT_DELAY = 15
MAX_ATTEMPTS = 10


class AWSProviderWithProxy(Provider):
    
    def __init__(self, region: str = None, proxy_config_file: str = None):
        super().__init__(region)
        self.current_proxy: Optional[ProxyInfo] = None
        
        # 初始化代理池
        if proxy_config_file:
            init_proxy_pool(proxy_config_file)
            logger.info(f"Initialized proxy pool from {proxy_config_file}")
        
        # 获取下一个可用代理
        self._rotate_proxy()

    def _rotate_proxy(self):
        """轮换到下一个可用代理"""
        proxy_pool = get_global_proxy_pool()
        self.current_proxy = proxy_pool.get_next_proxy()
        
        if self.current_proxy:
            logger.info(f"Switched to proxy: {self.current_proxy.host}:{self.current_proxy.port}")
        else:
            logger.warning("No proxy available, using direct connection")

    def _generate_proxy_user_data(self) -> str:
        """生成包含代理配置的user data脚本"""
        if not self.current_proxy:
            return ""
        
        proxy_url = self._format_proxy_url(self.current_proxy)
        
        user_data_script = f"""#!/bin/bash
# Configure system proxy
echo 'export http_proxy={proxy_url}' >> /etc/environment
echo 'export https_proxy={proxy_url}' >> /etc/environment
echo 'export HTTP_PROXY={proxy_url}' >> /etc/environment  
echo 'export HTTPS_PROXY={proxy_url}' >> /etc/environment

# Configure apt proxy
cat > /etc/apt/apt.conf.d/95proxy << EOF
Acquire::http::Proxy "{proxy_url}";
Acquire::https::Proxy "{proxy_url}";
EOF

# Configure chrome/chromium proxy
mkdir -p /etc/opt/chrome/policies/managed
cat > /etc/opt/chrome/policies/managed/proxy.json << EOF
{{
    "ProxyMode": "fixed_servers",
    "ProxyServer": "{self.current_proxy.host}:{self.current_proxy.port}"
}}
EOF

# Configure chromium proxy (Ubuntu default)
mkdir -p /etc/chromium/policies/managed
cat > /etc/chromium/policies/managed/proxy.json << EOF
{{
    "ProxyMode": "fixed_servers", 
    "ProxyServer": "{self.current_proxy.host}:{self.current_proxy.port}"
}}
EOF

# Configure firefox proxy - support multiple possible paths
for firefox_dir in /etc/firefox/policies /usr/lib/firefox/distribution/policies /etc/firefox-esr/policies; do
    if [ -d "$(dirname "$firefox_dir")" ]; then
        mkdir -p "$firefox_dir"
        cat > "$firefox_dir/policies.json" << EOF
{{
    "policies": {{
        "Proxy": {{
            "Mode": "manual",
            "HTTPProxy": "{self.current_proxy.host}:{self.current_proxy.port}",
            "HTTPSProxy": "{self.current_proxy.host}:{self.current_proxy.port}",
            "UseHTTPProxyForAllProtocols": true
        }}
    }}
}}
EOF
        break
    fi
done

# Reload environment variables
source /etc/environment

# Log proxy configuration
echo "$(date): Configured proxy {self.current_proxy.host}:{self.current_proxy.port}" >> /var/log/proxy-setup.log
"""
        
        return base64.b64encode(user_data_script.encode()).decode()

    def _format_proxy_url(self, proxy: ProxyInfo) -> str:
        """格式化代理URL"""
        if proxy.username and proxy.password:
            return f"{proxy.protocol}://{proxy.username}:{proxy.password}@{proxy.host}:{proxy.port}"
        else:
            return f"{proxy.protocol}://{proxy.host}:{proxy.port}"

    def start_emulator(self, path_to_vm: str, headless: bool, *args, **kwargs):
        logger.info("Starting AWS VM with proxy configuration...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            # 如果实例已经存在，直接启动
            ec2_client.start_instances(InstanceIds=[path_to_vm])
            logger.info(f"Instance {path_to_vm} is starting...")

            # Wait for the instance to be in the 'running' state
            waiter = ec2_client.get_waiter('instance_running')
            waiter.wait(InstanceIds=[path_to_vm], WaiterConfig={'Delay': WAIT_DELAY, 'MaxAttempts': MAX_ATTEMPTS})
            logger.info(f"Instance {path_to_vm} is now running.")

        except ClientError as e:
            logger.error(f"Failed to start the AWS VM {path_to_vm}: {str(e)}")
            raise

    def create_instance_with_proxy(self, image_id: str, instance_type: str, 
                                 security_groups: list, subnet_id: str) -> str:
        """创建带有代理配置的新实例"""
        ec2_client = boto3.client('ec2', region_name=self.region)
        
        user_data = self._generate_proxy_user_data()
        
        run_instances_params = {
            "MaxCount": 1,
            "MinCount": 1,
            "ImageId": image_id,
            "InstanceType": instance_type,
            "EbsOptimized": True,
            "NetworkInterfaces": [
                {
                    "SubnetId": subnet_id,
                    "AssociatePublicIpAddress": True,
                    "DeviceIndex": 0,
                    "Groups": security_groups
                }
            ]
        }
        
        if user_data:
            run_instances_params["UserData"] = user_data
            
        try:
            response = ec2_client.run_instances(**run_instances_params)
            instance_id = response['Instances'][0]['InstanceId']
            
            logger.info(f"Created new instance {instance_id} with proxy configuration")
            
            logger.info(f"Waiting for instance {instance_id} to be running...")
            ec2_client.get_waiter('instance_running').wait(InstanceIds=[instance_id])
            logger.info(f"Instance {instance_id} is ready.")

            try:
                instance_details = ec2_client.describe_instances(InstanceIds=[instance_id])
                instance = instance_details['Reservations'][0]['Instances'][0]
                public_ip = instance.get('PublicIpAddress', '')
                if public_ip:
                    vnc_url = f"http://{public_ip}:5910/vnc.html"
                    logger.info("="*80)
                    logger.info(f"🖥️  VNC Web Access URL: {vnc_url}")
                    logger.info(f"📡 Public IP: {public_ip}")
                    logger.info(f"🆔 Instance ID: {instance_id}")
                    if self.current_proxy:
                        logger.info(f"🌐 Proxy: {self.current_proxy.host}:{self.current_proxy.port}")
                    logger.info("="*80)
                    print(f"\n🌐 VNC Web Access URL: {vnc_url}")
                    if self.current_proxy:
                        print(f"🔄 Current Proxy: {self.current_proxy.host}:{self.current_proxy.port}")
                    print(f"📍 Please open the above address in the browser for remote desktop access\n")
            except Exception as e:
                logger.warning(f"Failed to get VNC address for instance {instance_id}: {e}")
            
            return instance_id
            
        except ClientError as e:
            logger.error(f"Failed to create instance with proxy: {str(e)}")
            if self.current_proxy:
                proxy_pool = get_global_proxy_pool()
                proxy_pool.mark_proxy_failed(self.current_proxy)
                self._rotate_proxy()
            raise

    def get_ip_address(self, path_to_vm: str) -> str:
        logger.info("Getting AWS VM IP address...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            response = ec2_client.describe_instances(InstanceIds=[path_to_vm])
            for reservation in response['Reservations']:
                for instance in reservation['Instances']:
                    private_ip_address = instance.get('PrivateIpAddress', '')
                    public_ip_address = instance.get('PublicIpAddress', '')

                    if public_ip_address:
                        vnc_url = f"http://{public_ip_address}:5910/vnc.html"
                        logger.info("="*80)
                        logger.info(f"🖥️  VNC Web Access URL: {vnc_url}")
                        logger.info(f"📡 Public IP: {public_ip_address}")
                        logger.info(f"🏠 Private IP: {private_ip_address}")
                        if self.current_proxy:
                            logger.info(f"🌐 Proxy: {self.current_proxy.host}:{self.current_proxy.port}")
                        logger.info("="*80)
                        print(f"\n🌐 VNC Web Access URL: {vnc_url}")
                        if self.current_proxy:
                            print(f"🔄 Current Proxy: {self.current_proxy.host}:{self.current_proxy.port}")
                        print(f"📍 Please open the above address in the browser for remote desktop access\n")
                    else:
                        logger.warning("No public IP address available for VNC access")
                    
                    return private_ip_address
            return ''
        except ClientError as e:
            logger.error(f"Failed to retrieve IP address for the instance {path_to_vm}: {str(e)}")
            raise

    def save_state(self, path_to_vm: str, snapshot_name: str):
        logger.info("Saving AWS VM state...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            image_response = ec2_client.create_image(InstanceId=path_to_vm, Name=snapshot_name)
            image_id = image_response['ImageId']
            logger.info(f"AMI {image_id} created successfully from instance {path_to_vm}.")
            return image_id
        except ClientError as e:
            logger.error(f"Failed to create AMI from the instance {path_to_vm}: {str(e)}")
            raise

    def revert_to_snapshot(self, path_to_vm: str, snapshot_name: str):
        logger.info(f"Reverting AWS VM to snapshot: {snapshot_name}...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            # Get original instance details for config.
            instance_details = ec2_client.describe_instances(InstanceIds=[path_to_vm])
            instance = instance_details['Reservations'][0]['Instances'][0]
            security_groups = [sg['GroupId'] for sg in instance['SecurityGroups']]
            subnet_id = instance['SubnetId']
            instance_type = instance['InstanceType']

            # Terminate the old instance. This is a non-blocking call.
            logger.info(f"Initiating termination for old instance {path_to_vm}...")
            ec2_client.terminate_instances(InstanceIds=[path_to_vm])
            logger.info(f"Old instance {path_to_vm} termination initiated.")

            # Rotate to a new proxy
            self._rotate_proxy()
            
            # Create a new instance
            new_instance_id = self.create_instance_with_proxy(
                snapshot_name, instance_type, security_groups, subnet_id
            )
            
            # Note: VNC address is displayed within create_instance_with_proxy
            logger.info(f"Successfully launched new instance {new_instance_id} for revert.")

            return new_instance_id

        except ClientError as e:
            logger.error(f"Failed to revert to snapshot {snapshot_name} for the instance {path_to_vm}: {str(e)}")
            raise

    def stop_emulator(self, path_to_vm, region=None):
        logger.info(f"Stopping AWS VM {path_to_vm}...")
        ec2_client = boto3.client('ec2', region_name=self.region)

        try:
            ec2_client.stop_instances(InstanceIds=[path_to_vm])
            waiter = ec2_client.get_waiter('instance_stopped')
            waiter.wait(InstanceIds=[path_to_vm], WaiterConfig={'Delay': WAIT_DELAY, 'MaxAttempts': MAX_ATTEMPTS})
            logger.info(f"Instance {path_to_vm} has been stopped.")
        except ClientError as e:
            logger.error(f"Failed to stop the AWS VM {path_to_vm}: {str(e)}")
            raise

    def get_current_proxy_info(self) -> Optional[dict]:
        """获取当前代理信息"""
        if self.current_proxy:
            return {
                'host': self.current_proxy.host,
                'port': self.current_proxy.port,
                'protocol': self.current_proxy.protocol,
                'failed_count': self.current_proxy.failed_count
            }
        return None

    def force_rotate_proxy(self):
        """强制轮换代理"""
        logger.info("Force rotating proxy...")
        if self.current_proxy:
            proxy_pool = get_global_proxy_pool()
            proxy_pool.mark_proxy_failed(self.current_proxy)
        self._rotate_proxy()

    def get_proxy_stats(self) -> dict:
        """获取代理池统计信息"""
        proxy_pool = get_global_proxy_pool()
        return proxy_pool.get_stats() 


import csv


# I want to write a function, reads a csv file, and get all the contents in the third column in the order of rows
def get_conference_city_in_order(env, config):
    # read the csv file
    csv_path = config['csv_path']
    print(f"Reading csv file from {csv_path}")
    with open(csv_path, 'r') as f:
        reader = csv.reader(f)
    # skip the header row
    next(reader)
    # get the third column in the order of rows
    conference_city_list = [row[2] for row in reader]
    return conference_city_list



"""
Stream manager for per-task progress messaging.

This module provides a global `stream_manager` singleton that manages
async message queues for task-based streaming. The singleton is async-safe
and should have its event loop configured via `set_loop()` during application
startup.
"""
import asyncio
from google.protobuf.timestamp_pb2 import Timestamp
from typing import Dict, Optional, AsyncGenerator
from dataclasses import dataclass
import logging

logger = logging.getLogger(__name__)


@dataclass
class StreamMessage:
    stage: str
    message: str
    timestamp: Timestamp


class StreamManager:
    """
    Manages in-memory async message queues for each task to stream progress.
    This class is async-safe.
    """

    def __init__(self, max_queue_size: int = 100):
        """
        Initialize a StreamManager that manages per-task in-memory async message queues.
        
        Parameters:
            max_queue_size (int): Maximum number of messages to keep per task queue; when a queue is full the oldest message will be dropped to make room for new messages.
        """
        self.task_queues: Dict[str, asyncio.Queue[Optional[StreamMessage]]] = {}
        self.max_queue_size = max_queue_size
        self._lock = asyncio.Lock()
        self.loop: Optional[asyncio.AbstractEventLoop] = None

    def set_loop(self, loop: asyncio.AbstractEventLoop):
        """
        Store the event loop used to schedule coroutines from non-async threads.
        
        Parameters:
            loop (asyncio.AbstractEventLoop): Event loop passed to asyncio.run_coroutine_threadsafe for thread-safe coroutine execution.
        """
        self.loop = loop

    def add_message_threadsafe(self, task_id: str, stage: str, message: str):
        """
        Enqueue a progress message for a task from a non-async thread in a thread-safe manner.
        
        If the manager's event loop has not been set, an error is logged and the message is not scheduled.
        
        Parameters:
            task_id (str): Identifier of the task to receive the message.
            stage (str): Stage label for the progress update.
            message (str): Text of the progress message.
        """
        if not self.loop:
            logger.error("StreamManager event loop not set. Cannot send message from thread.")
            return

        asyncio.run_coroutine_threadsafe(
            self.add_message(task_id, stage, message),
            self.loop
        )

    async def add_message(self, task_id: str, stage: str, message: str):
        """
        Enqueues a progress message for the given task; if the task's queue is full, drops the oldest message to make room.
        
        Parameters:
            task_id (str): Identifier of the task whose queue will receive the message.
            stage (str): Short stage name or label for the message.
            message (str): Human-readable progress message.
        """
        async with self._lock:
            q = self.task_queues.get(task_id)

        if q:
            timestamp = Timestamp()
            timestamp.GetCurrentTime()
            msg = StreamMessage(stage=stage, message=message, timestamp=timestamp)
            try:
                q.put_nowait(msg)
            except asyncio.QueueFull:
                logger.warning(f"Message queue for task {task_id} is full. Dropping oldest message.")
                # Drop the oldest message to make space for the new one
                q.get_nowait()
                q.put_nowait(msg)

        else:
            logger.warning(f"No message queue found for task {task_id}. Message not added.")

    async def get_message_stream(self, task_id: str) -> AsyncGenerator[StreamMessage, None]:
        """
        Provide an async generator that yields progress messages for the given task.
        
        If the task has no existing queue, one is created and registered. The generator yields StreamMessage objects produced for the task and terminates when a sentinel `None` is received, signaling end of stream.
        
        Parameters:
            task_id (str): Identifier of the task whose message stream to consume.
        
        Returns:
            AsyncGenerator[StreamMessage, None]: An async generator yielding `StreamMessage` instances for the task; iteration ends when a sentinel `None` is encountered.
        """
        async with self._lock:
            if task_id not in self.task_queues:
                self.task_queues[task_id] = asyncio.Queue(maxsize=self.max_queue_size)
                logger.info(f"Registered message queue for task {task_id} in get_message_stream.")
            q = self.task_queues[task_id]

        while True:
            message = await q.get()
            if message is None:  # Sentinel value indicates end of stream
                logger.info(f"End of stream for task {task_id}")
                break
            yield message

    async def register_task(self, task_id: str):
        """
        Create a per-task message queue if one does not already exist.
        
        This is idempotent: if a queue for the given task_id already exists, the call has no effect. The created queue uses the manager's configured max_queue_size and the operation is safe to call concurrently.
        
        Parameters:
            task_id (str): Unique identifier of the task to register a message queue for.
        """
        async with self._lock:
            if task_id not in self.task_queues:
                self.task_queues[task_id] = asyncio.Queue(maxsize=self.max_queue_size)
                logger.info(f"Registered message queue for task {task_id}")

    async def unregister_task(self, task_id: str):
        """Removes a task's message queue and signals end of stream."""
        q = None
        async with self._lock:
            if task_id in self.task_queues:
                q = self.task_queues.pop(task_id)
                logger.info(f"Unregistered message queue for task {task_id}")
        if q:
            try:
                # Put a sentinel value to unblock any consumers
                q.put_nowait(None)
            except asyncio.QueueFull:
                # If full, make space for sentinel
                try:
                    q.get_nowait()
                except asyncio.QueueEmpty:
                    pass
                # Retry put after making space or if queue became empty
                try:
                    q.put_nowait(None)
                except asyncio.QueueFull:
                    logger.error(f"Could not send sentinel for task {task_id}: queue still full after retry")


# Global instance to be used across the application
stream_manager = StreamManager()


#!/usr/bin/env python3
"""
Test script for task cancellation functionality in gRPC Agent service.
"""

import asyncio
import logging
import time
import uuid
import grpc
import os
from pathlib import Path

# Add the project root to Python path
project_root = Path(__file__).parent
os.environ["PYTHONPATH"] = str(project_root)

from gui_agents.proto import agent_pb2, agent_pb2_grpc
from gui_agents.proto.pb.agent_pb2 import LLMConfig, StageModelConfig, CommonConfig, Authorization, InstanceMode, Sandbox, SandboxOS

# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

class TaskCancellationTest:
    """Test class for task cancellation functionality"""

    def __init__(self, grpc_address='localhost:50051'):
        self.grpc_address = grpc_address
        self.channel = None
        self.stub = None

    async def connect(self):
        """Connect to gRPC server"""
        try:
            self.channel = grpc.aio.insecure_channel(self.grpc_address)
            grpc.grpc.channel_ready_future(self.channel).result(timeout=10)
            self.stub = agent_pb2_grpc.AgentStub(self.channel)
            logger.info(f"Connected to gRPC server at {self.grpc_address}")
            return True
        except Exception as e:
            logger.error(f"Failed to connect to gRPC server: {e}")
            return False

    async def disconnect(self):
        """Disconnect from gRPC server"""
        if self.channel:
            await self.channel.close()
            logger.info("Disconnected from gRPC server")

    def create_test_request(self, instruction="Open a calculator and wait for 30 seconds"):
        """Create a test request for agent instruction"""
        # Create a simple authorization for testing
        auth = Authorization(
            orgID="test-org",
            apiKey="test-key"
        )

        # Create a common config
        common_config = CommonConfig(
            id="test-task",
            backend="lybic",
            mode=InstanceMode.FAST,
            steps=10,
            authorizationInfo=auth
        )

        # Create a simple sandbox config
        sandbox = Sandbox(
            id="",
            name="test-sandbox",
            description="Test sandbox for cancellation",
            shapeName="beijing-2c-4g-cpu",
            os=SandboxOS.LINUX
        )

        # Create the request
        request = agent_pb2.RunAgentInstructionRequest(
            instruction=instruction,
            sandbox=sandbox,
            runningConfig=common_config
        )

        return request

    async def test_cancel_task_sync(self):
        """Test cancelling a task started synchronously"""
        logger.info("=== Testing Synchronous Task Cancellation ===")

        try:
            # Start a task synchronously
            request = self.create_test_request("Open calculator and wait 60 seconds")
            logger.info("Starting task synchronously...")

            task_id = None

            # Start the task and stream responses
            async for response in self.stub.RunAgentInstruction(request):
                if not task_id:
                    task_id = response.taskId
                    logger.info(f"Task started with ID: {task_id}")

                logger.info(f"Task update: {response.stage} - {response.message}")

                # Cancel after receiving a few messages
                if response.stage == "starting":
                    logger.info("Cancelling task...")
                    cancel_request = agent_pb2.CancelTaskRequest(taskId=task_id)
                    cancel_response = await self.stub.CancelTask(cancel_request)
                    logger.info(f"Cancel response: {cancel_response.success} - {cancel_response.message}")
                    break

            # Query task status to confirm cancellation
            if task_id:
                await asyncio.sleep(2)  # Give it time to process
                status_request = agent_pb2.QueryTaskStatusRequest(taskId=task_id)
                status_response = await self.stub.QueryTaskStatus(status_request)
                logger.info(f"Final task status: {status_response.status} - {status_response.message}")

            return True

        except Exception as e:
            logger.error(f"Error in synchronous task cancellation test: {e}")
            return False

    async def test_cancel_task_async(self):
        """Test cancelling a task started asynchronously"""
        logger.info("=== Testing Asynchronous Task Cancellation ===")

        try:
            # Start a task asynchronously
            request = self.create_test_request("Open notepad and wait 60 seconds")
            logger.info("Starting task asynchronously...")

            async_response = await self.stub.RunAgentInstructionAsync(request)
            task_id = async_response.taskId
            logger.info(f"Task started with ID: {task_id}")

            # Wait a moment then cancel
            await asyncio.sleep(2)
            logger.info("Cancelling task...")

            cancel_request = agent_pb2.CancelTaskRequest(taskId=task_id)
            cancel_response = await self.stub.CancelTask(cancel_request)
            logger.info(f"Cancel response: {cancel_response.success} - {cancel_response.message}")

            # Query task status to confirm cancellation
            await asyncio.sleep(2)
            status_request = agent_pb2.QueryTaskStatusRequest(taskId=task_id)
            status_response = await self.stub.QueryTaskStatus(status_request)
            logger.info(f"Final task status: {status_response.status} - {status_response.message}")

            return True

        except Exception as e:
            logger.error(f"Error in asynchronous task cancellation test: {e}")
            return False

    async def test_cancel_nonexistent_task(self):
        """Test cancelling a task that doesn't exist"""
        logger.info("=== Testing Non-existent Task Cancellation ===")

        try:
            fake_task_id = str(uuid.uuid4())
            logger.info(f"Attempting to cancel non-existent task: {fake_task_id}")

            cancel_request = agent_pb2.CancelTaskRequest(taskId=fake_task_id)
            cancel_response = await self.stub.CancelTask(cancel_request)
            logger.info(f"Cancel response: {cancel_response.success} - {cancel_response.message}")

            if not cancel_response.success and "not found" in cancel_response.message.lower():
                logger.info("✓ Correctly handled non-existent task")
                return True
            else:
                logger.error("✗ Did not handle non-existent task correctly")
                return False

        except Exception as e:
            logger.error(f"Error in non-existent task cancellation test: {e}")
            return False

    async def test_cancel_completed_task(self):
        """Test cancelling a task that has already completed"""
        logger.info("=== Testing Completed Task Cancellation ===")

        try:
            # Start a simple task that completes quickly
            request = self.create_test_request("Just say hello and finish")
            logger.info("Starting simple task...")

            task_id = None

            # Let the task complete
            async for response in self.stub.RunAgentInstruction(request):
                if not task_id:
                    task_id = response.taskId
                    logger.info(f"Task started with ID: {task_id}")

                logger.info(f"Task update: {response.stage} - {response.message}")

            # Wait for task to complete
            await asyncio.sleep(2)

            # Try to cancel the completed task
            if task_id:
                logger.info("Attempting to cancel completed task...")
                cancel_request = agent_pb2.CancelTaskRequest(taskId=task_id)
                cancel_response = await self.stub.CancelTask(cancel_request)
                logger.info(f"Cancel response: {cancel_response.success} - {cancel_response.message}")

                if not cancel_response.success and ("finished" in cancel_response.message.lower() or "completed" in cancel_response.message.lower()):
                    logger.info("✓ Correctly handled completed task")
                    return True
                else:
                    logger.error("✗ Did not handle completed task correctly")
                    return False

            return False

        except Exception as e:
            logger.error(f"Error in completed task cancellation test: {e}")
            return False

    async def run_all_tests(self):
        """Run all cancellation tests"""
        logger.info("Starting task cancellation tests...")

        if not await self.connect():
            logger.error("Failed to connect to gRPC server. Make sure the server is running.")
            return False

        try:
            results = []

            # Run all tests
            results.append(await self.test_cancel_nonexistent_task())
            await asyncio.sleep(1)

            results.append(await self.test_cancel_task_async())
            await asyncio.sleep(1)

            results.append(await self.test_cancel_task_sync())
            await asyncio.sleep(1)

            results.append(await self.test_cancel_completed_task())

            # Summary
            passed = sum(results)
            total = len(results)
            logger.info(f"\n=== Test Summary ===")
            logger.info(f"Passed: {passed}/{total}")

            if passed == total:
                logger.info("✓ All tests passed!")
                return True
            else:
                logger.error(f"✗ {total - passed} tests failed")
                return False

        finally:
            await self.disconnect()

async def main():
    """Main function to run the tests"""
    test = TaskCancellationTest()
    success = await test.run_all_tests()

    if success:
        logger.info("Task cancellation functionality test completed successfully!")
        exit(0)
    else:
        logger.error("Task cancellation functionality test failed!")
        exit(1)

if __name__ == "__main__":
    asyncio.run(main())


# pytest -s tests/test_hardware_interface.py
import platform
from time import sleep
import time
from gui_agents.agents.hardware_interface import HardwareInterface
from gui_agents.agents.Action import (
    Click, Hotkey, Screenshot, Wait, TypeText
)



current_platform = platform.system().lower()

# -------------------------------------------------------------
dry_run = False          # ← True 时只打印，不执行
backend_kwargs = dict(   # 传给 PyAutoGUIBackend 的额外参数
    platform=current_platform,    # "darwin" / "linux" / "win32"，自动检测可删掉
    default_move_duration=0.05,
)

# # 1. 构建动作序列
plan = [
    Click(x=10, y=300, element_description=''),
    # TypeText(text="Hello Action!"),
    # Wait(time=0.5),
    # Drag(start=(400, 300), end=(800, 300), hold_keys=[],
    #      starting_description="拖起点", ending_description="拖终点"),
    # Hotkey(keys=["ctrl", "s"]),
    # Open(app_or_filename='maps')
    # Screenshot()
    # Hotkey(keys=["command", "space"]),
    # TypeText(text="测试", element_description="", press_enter=True)
]

plan = [
    # {'type': 'Hotkey', 'keys': ['command', 'space'], 'duration': 80}
    # {
    #     "type": "TouchTap",
    #     "x": 130,
    #     "y": 478,
    # },
    # {
    #     "type": "TouchSwipe",
    #     "x": 540,
    #     "y": 960,
    #     "direction": "right",
    #     "distance": 300,
    # },
    {
        "type": "TouchDrag",
        "startX": 130,
        "startY": 180,
        "endX": 335,
        "endY": 787,
    },
    # {
    #     "type": "TypeText",
    #     "text": "hello",
    # },
    # {'type': 'Click', 'xy': [154, 64], 'element_description': 'The Chromium browser icon on the desktop, which is a circular icon with red, green, yellow, and blue colors', 'num_clicks': 2, 'button_type': 'left', 'hold_keys': []}
]

# 2. 创建硬件接口
hwi = HardwareInterface(backend="lybic_mobile", **backend_kwargs)
# hwi = HardwareInterface(backend="pyautogui", **backend_kwargs)

# 3. 执行
if dry_run:
    print("Dry-run 模式，仅打印 Action：")
    for a in plan:
        print(a)
else:
    print("开始执行 Action 序列…")
    # time.sleep(5)
    # hwi.dispatch(plan)
    res = hwi.dispatchDict(plan)
    # print(res)

    print("执行完毕")
